from typing import List, Union
import logging
import grpc
import time
from csle_common.dao.emulation_config.emulation_env_config import EmulationEnvConfig
from csle_common.dao.emulation_config.node_container_config import NodeContainerConfig
from csle_common.dao.emulation_config.five_g_cu_managers_info import FiveGCUManagersInfo
import csle_common.constants.constants as constants
import csle_collector.constants.constants as csle_collector_constants
import csle_collector.five_g_cu_manager.five_g_cu_manager_pb2_grpc
import csle_collector.five_g_cu_manager.five_g_cu_manager_pb2
import csle_collector.five_g_cu_manager.five_g_cu_manager_util
import csle_collector.five_g_cu_manager.query_five_g_cu_manager
from csle_common.util.emulation_util import EmulationUtil


class FiveGCUController:
    """
    Class controlling 5G cus running on nodes in the emulations, as well as 5G CU managers
    """

    @staticmethod
    def start_five_g_cu_managers(emulation_env_config: EmulationEnvConfig, physical_server_ip: str,
                                 logger: logging.Logger) -> None:
        """
        Utility method for starting 5G CU managers

        :param emulation_env_config: the emulation env config
        :param physical_server_ip: the ip of the physical server
        :param logger: the logger to use for logging
        :return: None
        """
        for c in emulation_env_config.containers_config.containers:
            if c.physical_host_ip != physical_server_ip:
                continue
            for ids_image in constants.CONTAINER_IMAGES.FIVE_G_CU_IMAGES:
                if ids_image in c.name:
                    FiveGCUController.start_five_g_cu_manager(emulation_env_config=emulation_env_config,
                                                              ip=c.docker_gw_bridge_ip, logger=logger)

    @staticmethod
    def start_five_g_cu_manager(emulation_env_config: EmulationEnvConfig, ip: str, logger: logging.Logger) -> None:
        """
        Utility method for starting the 5G CU manager on a specific container

        :param emulation_env_config: the emulation env config
        :param ip: the ip of the container
        :param logger: the logger to use for logging
        :return: None
        """
        # Check if the manager is already running
        status = None
        not_running = False
        try:
            status = FiveGCUController.get_five_g_cu_status_by_ip_and_port(
                ip=ip, port=emulation_env_config.five_g_config.five_g_cu_manager_port, timeout=5)
        except Exception:
            not_running = True
        status_str = ""
        if status is None:
            not_running = True
        else:
            status_str = f"cu_running: {status.cu_running}, ip: {status.ip}"
        if not_running:
            logger.info(f"Starting 5G CU manager on node {ip}")

            # Connect
            EmulationUtil.connect_admin(emulation_env_config=emulation_env_config, ip=ip)

            # Stop old background job if running
            cmd = (constants.COMMANDS.SUDO + constants.COMMANDS.SPACE_DELIM + constants.COMMANDS.PKILL +
                   constants.COMMANDS.SPACE_DELIM + constants.TRAFFIC_COMMANDS.FIVE_G_CU_MANAGER_FILE_NAME)
            o, e, _ = EmulationUtil.execute_ssh_cmd(
                cmd=cmd, conn=emulation_env_config.get_connection(ip=ip))

            # Start the 5G cu manager
            cmd = constants.COMMANDS.START_FIVE_G_CU_MANAGER.format(
                emulation_env_config.five_g_config.five_g_cu_manager_port,
                emulation_env_config.five_g_config.five_g_cu_manager_log_dir,
                emulation_env_config.five_g_config.five_g_cu_manager_log_file,
                emulation_env_config.five_g_config.five_g_cu_manager_max_workers)
            o, e, _ = EmulationUtil.execute_ssh_cmd(cmd=cmd, conn=emulation_env_config.get_connection(ip=ip))
            time.sleep(2)
        else:
            logger.info(f"5G CU manager was already running on node {ip}. Status: {status_str}")

    @staticmethod
    def stop_five_g_cu_managers(emulation_env_config: EmulationEnvConfig, physical_server_ip: str,
                                logger: logging.Logger) -> None:
        """
        Utility method for stopping 5G CU managers

        :param emulation_env_config: the emulation env config
        :param physical_server_ip: the IP of the physical host
        :param logger: the logger to use for logging
        :return: None
        """
        for c in emulation_env_config.containers_config.containers:
            if c.physical_host_ip != physical_server_ip:
                continue
            for ids_image in constants.CONTAINER_IMAGES.FIVE_G_CU_IMAGES:
                if ids_image in c.name:
                    FiveGCUController.stop_five_g_cu_manager(emulation_env_config=emulation_env_config,
                                                             ip=c.docker_gw_bridge_ip, logger=logger)

    @staticmethod
    def stop_five_g_cu_manager(emulation_env_config: EmulationEnvConfig, ip: str, logger: logging.Logger) -> None:
        """
        Utility method for stopping a 5G CU manager with a specific IP

        :param emulation_env_config: the emulation env config
        :param ip: the ip of the container
        :param logger: the logger to use for logging
        :return: None
        """
        # Connect
        EmulationUtil.connect_admin(emulation_env_config=emulation_env_config, ip=ip)

        logger.info(f"Stopping 5G CU manager on node {ip}")

        cmd = (constants.COMMANDS.SUDO + constants.COMMANDS.SPACE_DELIM + constants.COMMANDS.PKILL +
               constants.COMMANDS.SPACE_DELIM + constants.TRAFFIC_COMMANDS.FIVE_G_CU_MANAGER_FILE_NAME)
        o, e, _ = EmulationUtil.execute_ssh_cmd(cmd=cmd, conn=emulation_env_config.get_connection(ip=ip))
        time.sleep(2)

    @staticmethod
    def get_five_g_cu_manager_statuses(
            emulation_env_config: EmulationEnvConfig, physical_server_ip: str, logger: logging.Logger) \
            -> List[csle_collector.five_g_cu_manager.five_g_cu_manager_pb2.FiveGCUStatusDTO]:
        """
        A method that sends a request to the 5G CU manager on every container to get the status of the
        5G CU

        :param emulation_env_config: the emulation config
        :param physical_server_ip: the IP of the physical server
        :param logger: the logger to use for logging
        :return: List of monitor thread statuses
        """
        statuses = []
        FiveGCUController.start_five_g_cu_managers(emulation_env_config=emulation_env_config,
                                                   physical_server_ip=physical_server_ip, logger=logger)

        for c in emulation_env_config.containers_config.containers:
            if c.physical_host_ip != physical_server_ip:
                continue
            for ids_image in constants.CONTAINER_IMAGES.FIVE_G_CU_IMAGES:
                if ids_image in c.name:
                    status = FiveGCUController.get_five_g_cu_status_by_ip_and_port(
                        port=emulation_env_config.five_g_config.five_g_cu_manager_port,
                        ip=c.docker_gw_bridge_ip)
                    statuses.append(status)
        return statuses

    @staticmethod
    def get_five_g_cu_managers_ips(emulation_env_config: EmulationEnvConfig) -> List[str]:
        """
        A method that extracts the IPs of the 5G CU managers in a given emulation

        :param emulation_env_config: the emulation env config
        :return: the list of IP addresses
        """
        ips = []
        for c in emulation_env_config.containers_config.containers:
            for ids_image in constants.CONTAINER_IMAGES.FIVE_G_CU_IMAGES:
                if ids_image in c.name:
                    ips.append(c.docker_gw_bridge_ip)
        return ips

    @staticmethod
    def get_five_g_cu_managers_ports(emulation_env_config: EmulationEnvConfig) -> List[int]:
        """
        A method that extracts the ports of the FiveGCUManagers in a given emulation

        :param emulation_env_config: the emulation env config
        :return: the list of ports
        """
        ports = []
        for c in emulation_env_config.containers_config.containers:
            for ids_image in constants.CONTAINER_IMAGES.FIVE_G_CU_IMAGES:
                if ids_image in c.name:
                    ports.append(emulation_env_config.five_g_config.five_g_cu_manager_port)
        return ports

    @staticmethod
    def get_five_g_cu_status_by_ip_and_port(
            port: int, ip: str, timeout: int = csle_collector_constants.GRPC.TIMEOUT_SECONDS) \
            -> csle_collector.five_g_cu_manager.five_g_cu_manager_pb2.FiveGCUStatusDTO:
        """
        A method that sends a request to the 5G Cu manager with a specific port and ip
        to get the status of the 5G CU

        :param port: the port of the FiveGCUManager
        :param ip: the ip of the FiveGCUManager
        :param timeout: the timeout of the GRPC query
        :return: the status of the FiveGCUManager
        """
        with grpc.insecure_channel(f'{ip}:{port}', options=constants.GRPC_SERVERS.GRPC_OPTIONS) as channel:
            stub = csle_collector.five_g_cu_manager.five_g_cu_manager_pb2_grpc.FiveGCUManagerStub(channel)
            status = \
                csle_collector.five_g_cu_manager.query_five_g_cu_manager.get_five_g_cu_status(
                    stub=stub, timeout=timeout)
            return status

    @staticmethod
    def start_five_g_cus(emulation_env_config: EmulationEnvConfig, physical_server_ip: str,
                         logger: logging.Logger) -> None:
        """
        Utility method for starting the 5G CUs of a specific execution

        :param emulation_env_config: the emulation env config
        :param physical_server_ip: the ip of the physical server
        :param logger: the logger to use for logging
        :return: None
        """
        for c in emulation_env_config.containers_config.containers:
            if c.physical_host_ip != physical_server_ip:
                continue
            for ids_image in constants.CONTAINER_IMAGES.FIVE_G_CU_IMAGES:
                if ids_image in c.name:
                    FiveGCUController.start_five_g_cu(emulation_env_config=emulation_env_config,
                                                      ip=c.docker_gw_bridge_ip, logger=logger)

    @staticmethod
    def start_five_g_cu(emulation_env_config: EmulationEnvConfig, ip: str, logger: logging.Logger) \
            -> csle_collector.five_g_cu_manager.five_g_cu_manager_pb2.FiveGCUStatusDTO:
        """
        Utility method for starting the 5G CU on a specific container

        :param emulation_env_config: the emulation env config
        :param ip: the ip of the container
        :param logger: the logger to use for logging
        :return: None
        """
        logger.info(f"Starting the 5G CU on container with ip {ip} in execution {emulation_env_config.execution_id} "
                    f"of emulation: {emulation_env_config.name}")
        port = emulation_env_config.five_g_config.five_g_cu_manager_port
        with grpc.insecure_channel(f'{ip}:{port}', options=constants.GRPC_SERVERS.GRPC_OPTIONS) as channel:
            stub = csle_collector.five_g_cu_manager.five_g_cu_manager_pb2_grpc.FiveGCUManagerStub(channel)
            status = csle_collector.five_g_cu_manager.query_five_g_cu_manager.start_five_g_cu(stub=stub)
            return status

    @staticmethod
    def stop_five_g_cus(emulation_env_config: EmulationEnvConfig, physical_server_ip: str,
                        logger: logging.Logger) -> None:
        """
        Utility method for stopping the 5G CUs of a specific execution

        :param emulation_env_config: the emulation env config
        :param physical_server_ip: the ip of the physical server
        :param logger: the logger to use for logging
        :return: None
        """
        for c in emulation_env_config.containers_config.containers:
            if c.physical_host_ip != physical_server_ip:
                continue
            for ids_image in constants.CONTAINER_IMAGES.FIVE_G_CU_IMAGES:
                if ids_image in c.name:
                    FiveGCUController.stop_five_g_cu(emulation_env_config=emulation_env_config,
                                                     ip=c.docker_gw_bridge_ip, logger=logger)

    @staticmethod
    def stop_five_g_cu(emulation_env_config: EmulationEnvConfig, ip: str, logger: logging.Logger) \
            -> csle_collector.five_g_cu_manager.five_g_cu_manager_pb2.FiveGCUStatusDTO:
        """
        Utility method for stopping the 5G cu on a specific container

        :param emulation_env_config: the emulation env config
        :param ip: the ip of the container
        :param logger: the logger to use for logging
        :return: None
        """
        logger.info(f"Stopping the 5G CU on container with ip {ip} in execution {emulation_env_config.execution_id} "
                    f"of emulation: {emulation_env_config.name}")
        port = emulation_env_config.five_g_config.five_g_cu_manager_port
        with grpc.insecure_channel(f'{ip}:{port}', options=constants.GRPC_SERVERS.GRPC_OPTIONS) as channel:
            stub = csle_collector.five_g_cu_manager.five_g_cu_manager_pb2_grpc.FiveGCUManagerStub(channel)
            status = csle_collector.five_g_cu_manager.query_five_g_cu_manager.stop_five_g_cu(stub=stub)
            return status

    @staticmethod
    def get_five_g_cu_managers_info(emulation_env_config: EmulationEnvConfig, active_ips: List[str],
                                    logger: logging.Logger, physical_server_ip: str) -> FiveGCUManagersInfo:
        """
        Extracts the information of the 5G CU managers for a given emulation

        :param emulation_env_config: the configuration of the emulation
        :param active_ips: list of active IPs
        :param physical_server_ip: the IP of the physical server
        :param logger: the logger to use for logging
        :return: a DTO with the status of the 5G CU managers
        """
        five_g_cu_managers_ips = FiveGCUController.get_five_g_cu_managers_ips(
            emulation_env_config=emulation_env_config)
        five_g_cu_managers_ports = FiveGCUController.get_five_g_cu_managers_ports(
            emulation_env_config=emulation_env_config)
        five_g_cu_managers_statuses = []
        five_g_cu_managers_running = []
        for ip in five_g_cu_managers_ips:
            if ip not in active_ips or not EmulationUtil.physical_ip_match(
                    emulation_env_config=emulation_env_config, ip=ip, physical_host_ip=physical_server_ip):
                continue
            running = False
            status = None
            try:
                status = FiveGCUController.get_five_g_cu_status_by_ip_and_port(
                    port=emulation_env_config.five_g_config.five_g_cu_manager_port, ip=ip)
                running = True
            except Exception as e:
                logger.debug(
                    f"Could not fetch 5G CU manager status on IP:{ip}, error: {str(e)}, {repr(e)}")
            if status is not None:
                five_g_cu_managers_statuses.append(status)
            else:
                util = csle_collector.five_g_cu_manager.five_g_cu_manager_util.FiveGCUManagerUtil
                five_g_cu_managers_statuses.append(util.five_g_cu_status_dto_empty())
            five_g_cu_managers_running.append(running)
        execution_id = emulation_env_config.execution_id
        emulation_name = emulation_env_config.name
        five_g_cu_manager_info_dto = FiveGCUManagersInfo(
            five_g_cu_managers_running=five_g_cu_managers_running, ips=five_g_cu_managers_ips,
            ports=five_g_cu_managers_ports, execution_id=execution_id, emulation_name=emulation_name,
            five_g_cu_managers_statuses=five_g_cu_managers_statuses)
        return five_g_cu_manager_info_dto

    @staticmethod
    def init_five_g_cus(emulation_env_config: EmulationEnvConfig, physical_server_ip: str,
                        logger: logging.Logger) -> None:
        """
        Utility method for initializing the 5G CUs of a specific execution

        :param emulation_env_config: the emulation env config
        :param physical_server_ip: the ip of the physical server
        :param logger: the logger to use for logging
        :return: None
        """
        for c in emulation_env_config.containers_config.containers:
            if c.physical_host_ip != physical_server_ip:
                continue
            for ids_image in constants.CONTAINER_IMAGES.FIVE_G_CU_IMAGES:
                if ids_image in c.name:
                    FiveGCUController.init_five_g_cu(emulation_env_config=emulation_env_config,
                                                     container=c, logger=logger)

    @staticmethod
    def init_five_g_cu(emulation_env_config: EmulationEnvConfig, container: NodeContainerConfig,
                       logger: logging.Logger) \
            -> Union[csle_collector.five_g_cu_manager.five_g_cu_manager_pb2.FiveGCUStatusDTO, None]:
        """
        Utility method for initializing the 5G CU on a specific container

        :param emulation_env_config: the emulation env config
        :param container: the container
        :param logger: the logger to use for logging
        :return: None
        """
        logger.info(f"Initializing the 5G CU on container with ip {container.docker_gw_bridge_ip} "
                    f"in execution {emulation_env_config.execution_id} "
                    f"of emulation: {emulation_env_config.name}")
        port = emulation_env_config.five_g_config.five_g_cu_manager_port
        core_backhaul_ip = emulation_env_config.five_g_config.core_backhaul_ip
        cu_backhaul_ip = ""
        for cu_ip in emulation_env_config.five_g_config.cu_backhaul_ips:
            if cu_ip in container.get_ips():
                cu_backhaul_ip = cu_ip
        if cu_backhaul_ip == "":
            return None
        cu_fronthaul_ip = ""
        for cu_ip in emulation_env_config.five_g_config.cu_fronthaul_ips:
            if cu_ip in container.get_ips():
                cu_fronthaul_ip = cu_ip
        if cu_fronthaul_ip == "":
            return None
        with grpc.insecure_channel(f'{container.docker_gw_bridge_ip}:{port}',
                                   options=constants.GRPC_SERVERS.GRPC_OPTIONS) as channel:
            stub = csle_collector.five_g_cu_manager.five_g_cu_manager_pb2_grpc.FiveGCUManagerStub(channel)
            status = csle_collector.five_g_cu_manager.query_five_g_cu_manager.init_five_g_cu(
                core_backhaul_ip=core_backhaul_ip, cu_backhaul_ip=cu_backhaul_ip, cu_fronthaul_ip=cu_fronthaul_ip,
                stub=stub)
            return status

    @staticmethod
    def start_five_g_cu_monitor_threads(emulation_env_config: EmulationEnvConfig, physical_server_ip: str,
                                        logger: logging.Logger) -> None:
        """
        A method that sends a request to the 5G CUManager on every container
        to start the CU manager and the monitor thread

        :param emulation_env_config: the emulation env config
        :param physical_server_ip: the ip of the physical server
        :param logger: the logger to use for logging
        :return: None
        """
        for c in emulation_env_config.containers_config.containers:
            if c.physical_host_ip != physical_server_ip:
                continue
            for container_image in constants.CONTAINER_IMAGES.FIVE_G_CU_IMAGES:
                if container_image in c.name:
                    FiveGCUController.start_five_g_cu_monitor_thread(emulation_env_config=emulation_env_config,
                                                                     ip=c.docker_gw_bridge_ip, logger=logger)

    @staticmethod
    def start_five_g_cu_monitor_thread(emulation_env_config: EmulationEnvConfig, ip: str, logger: logging.Logger) \
            -> None:
        """
        A method that sends a request to the FiveGCUManager on a specific IP
        to start the 5G CU monitor thread

        :param emulation_env_config: the emulation env config
        :param ip: IP of the container
        :param logger: the logger to use for logging
        :return: None
        """
        cu_status_dto = FiveGCUController.get_five_g_cu_status_by_ip_and_port(
            ip=ip, port=emulation_env_config.five_g_config.five_g_cu_manager_port)
        if not cu_status_dto.monitor_running:
            logger.info(f"5G CU monitor thread is not running on {ip}, starting it.")
            # Open a gRPC session
            with grpc.insecure_channel(
                    f'{ip}:{emulation_env_config.five_g_config.five_g_cu_manager_port}',
                    options=constants.GRPC_SERVERS.GRPC_OPTIONS) as channel:
                stub = csle_collector.five_g_cu_manager.five_g_cu_manager_pb2_grpc.FiveGCUManagerStub(channel)
                csle_collector.five_g_cu_manager.query_five_g_cu_manager.start_five_g_cu_monitor(
                    stub=stub, kafka_ip=emulation_env_config.kafka_config.container.get_ips()[0],
                    kafka_port=emulation_env_config.kafka_config.kafka_port,
                    time_step_len_seconds=emulation_env_config.kafka_config.time_step_len_seconds)

    @staticmethod
    def stop_five_g_cu_monitor_threads(emulation_env_config: EmulationEnvConfig, logger: logging.Logger,
                                       physical_host_ip: str) -> None:
        """
        A method that sends a request to the 5G CU on every container to stop the monitor threads

        :param emulation_env_config: the emulation env config
        :param physical_host_ip: the IP of the physical host
        :param logger: the logger to use for logging
        :return: None
        """
        for c in emulation_env_config.containers_config.containers:
            if c.physical_host_ip != physical_host_ip:
                continue
            for container_image in constants.CONTAINER_IMAGES.FIVE_G_CU_IMAGES:
                if container_image in c.name:
                    FiveGCUController.stop_five_g_cu_monitor_thread(emulation_env_config=emulation_env_config,
                                                                    ip=c.docker_gw_bridge_ip, logger=logger)

    @staticmethod
    def stop_five_g_cu_monitor_thread(emulation_env_config: EmulationEnvConfig, ip: str, logger: logging.Logger) \
            -> None:
        """
        A method that sends a request to the 5G CU Manager on a specific container to stop the monitor thread

        :param emulation_env_config: the emulation env config
        :param ip: the IP of the container
        :param logger: the logger to use for logging
        :return: None
        """
        # Open a gRPC session
        with grpc.insecure_channel(f'{ip}:{emulation_env_config.five_g_config.five_g_cu_manager_port}',
                                   options=constants.GRPC_SERVERS.GRPC_OPTIONS) as channel:
            stub = csle_collector.five_g_cu_manager.five_g_cu_manager_pb2_grpc.FiveGCUManagerStub(channel)
            logger.info(f"Stopping the 5G CU monitor thread on {ip}.")
            csle_collector.five_g_cu_manager.query_five_g_cu_manager.stop_five_g_cu_monitor(stub=stub)



"""
Transport protocols in CSLE
"""
from enum import IntEnum


class TransportProtocol(IntEnum):
    """
    Enum representing the different transport protocols in the network.
    """
    TCP = 0
    UDP = 1

    @staticmethod
    def _from_str(protocol_str: str) -> "TransportProtocol":
        """
        Creates the object from a string
        :param protocol_str: the string to create the object from
        :return: the created object
        """

        if protocol_str.lower() == "tcp":
            return TransportProtocol.TCP
        elif protocol_str.lower() == "udp":
            return TransportProtocol.UDP
        else:
            raise ValueError("Protocol string:{} not recognized".format(protocol_str))



from typing import Tuple, Dict, Any
import io
import csv
import json
import os
import zipfile
import csle_common.constants.constants as constants
from csle_common.logging.log import Logger
from csle_common.metastore.metastore_facade import MetastoreFacade


class ExportUtil:
    """
    Class with utility functions for exporting data from the metastore
    """

    @staticmethod
    def zipdir(dir_path: str, file_path: str) -> None:
        """
        Creates a zip file of a given directory

        :param dir_path: the path to the directory to zip
        :param file_path: the full path of the resulting zip file
        :return: None
        """
        Logger.__call__().get_logger().info(f"Zipping directory: {dir_path} to file: {file_path}")
        with zipfile.ZipFile(file_path, 'w', zipfile.ZIP_DEFLATED) as zipf:
            for root, dirs, files in os.walk(dir_path):
                for i, file in enumerate(files):
                    Logger.__call__().get_logger().info(f"Processing file {i + 1}/{len(files)}, file name: {file}")
                    zipf.write(os.path.join(root, file),
                               os.path.relpath(os.path.join(root, file),
                                               os.path.join(dir_path, '..')))

    @staticmethod
    def get_dir_size_gb(dir_path: str = '.') -> float:
        """
        Utility method to calculate the size of a file directory in gb

        :param dir_path: the path to the directory
        :return: the size of the directory in GB
        """
        total = 0.0
        with os.scandir(dir_path) as it:
            for entry in it:
                if entry.is_file():
                    total += entry.stat().st_size
                elif entry.is_dir():
                    total += ExportUtil.get_dir_size_gb(entry.path)
        return round((float(total) / 1000000000), 2)

    @staticmethod
    def export_emulation_traces_to_disk_json(num_traces_per_file: int, output_dir: str, zip_file_output: str,
                                             max_num_traces: int, added_by: str = "unknown", offset: int = 0,
                                             file_start_id: int = 1, emulation_traces_ids=None) -> None:
        """
        Exports emulation traces from the metastore to disk

        :param num_traces_per_file: the number of traces per file in the output directory
        :param output_dir: the output directory
        :param zip_file_output: the compressed zip file path
        :param max_num_traces: maximum number of traces
        :param added_by: the person who added the dataset
        :param offset: the trace id offset
        :param file_start_id: the id of the first file to write
        :param emulation_traces_ids: the ids of the traces to export
        :return: None
        """
        Logger.__call__().get_logger().info(f"Exporting emulation traces to disk (json), output dir: {output_dir}, "
                                            f"output zip file: {zip_file_output}, "
                                            f"num traces per file: {num_traces_per_file}")
        if not os.path.exists(output_dir):
            os.makedirs(output_dir)
        if emulation_traces_ids is None:
            emulation_traces_ids = MetastoreFacade.list_emulation_traces_ids()
            emulation_traces_ids = emulation_traces_ids[offset:]
        if len(emulation_traces_ids) > max_num_traces:
            emulation_traces_ids = emulation_traces_ids[0:max_num_traces]
        traces = []
        file_id = file_start_id
        file_name = f"{file_id}.json"
        columns = ""
        last_export = 0
        num_attributes_per_time_step = -1
        schema = None
        for i, id_obj in enumerate(emulation_traces_ids):
            Logger.__call__().get_logger().info(f"Reading trace {i}/{len(emulation_traces_ids)} from the metastore, "
                                                f"trace id: {id_obj[0]}")
            tr = MetastoreFacade.get_emulation_trace(id=id_obj[0])
            if tr is None:
                raise ValueError(f"Could not find an emulation trace with id: {id_obj[0]}")
            if num_attributes_per_time_step == -1:
                num_attributes_per_time_step = tr.num_attributes_per_time_step()
            if schema is None:
                schema = {
                    constants.METADATA_STORE.TRACES_PROPERTY: [
                        tr.schema().to_dict()
                    ]
                }

            traces.append(tr.to_dict())
            if i > 0 and ((i % num_traces_per_file == 0) or i == (len(emulation_traces_ids) - 1)):
                Logger.__call__().get_logger().info(f"Exporting traces {last_export + 1}-{i} to file: {file_name}")
                traces_dict = {constants.METADATA_STORE.TRACES_PROPERTY: traces}
                traces_str = json.dumps(traces_dict, indent=4, sort_keys=True)
                with io.open(f"{output_dir}{constants.COMMANDS.SLASH_DELIM}{file_name}", 'w', encoding='utf-8') as f:
                    f.write(traces_str)
                file_id += 1
                file_name = f"{file_id}.json"
                traces = []
                last_export = i
        file_format = "json"
        num_traces = len(emulation_traces_ids)
        with io.open(f"{output_dir}{constants.COMMANDS.SLASH_DELIM}{constants.DATASETS.METADATA_FILE_NAME}", 'w',
                     encoding='utf-8') as f:
            metadata_dict: Dict[str, Any] = {}
            metadata_dict[constants.DATASETS.FILE_FORMAT_PROPERTY] = file_format
            metadata_dict[constants.DATASETS.NUM_TRACES_PROPERTY] = num_traces
            metadata_dict[constants.DATASETS.NUM_ATTRIBUTES_PER_TIME_STEP_PROPERTY] = num_attributes_per_time_step
            metadata_dict[constants.DATASETS.SCHEMA_PROPERTY] = schema
            metadata_dict[constants.DATASETS.NUM_TRACES_PER_FILE_PROPERTY] = num_traces_per_file
            metadata_dict[constants.DATASETS.ADDED_BY_PROPERTY] = added_by
            metadata_dict[constants.DATASETS.COLUMNS_PROPERTY] = columns
            f.write(json.dumps(metadata_dict, indent=4, sort_keys=True))
        ExportUtil.zipdir(dir_path=output_dir, file_path=zip_file_output)
        Logger.__call__().get_logger().info(f"Export of emulation traces to disk complete, "
                                            f"output dir:{output_dir}, output zip file: {zip_file_output}")

    @staticmethod
    def extract_emulation_traces_dataset_metadata(dir_path: str, zip_file_path: str) \
            -> Tuple[int, float, float, str, int, str, int, int, str, str]:
        """
        Extracts metadata of a traces dataset stored on disk

        :param dir_path: the path to the directory where the traces dataset is stored
        :param zip_file_path: the path to the compressed zipfile of the dataset
        :return: num_files, dir_size_uncompressed_gb, file_format, num_traces, schema, num_traces_per_file,
                 num_attributes_per_time_step, added_by, columns
        """
        file_format = "unknown"
        added_by = "unknown"
        num_traces = -1
        schema = ""
        columns = ""
        num_traces_per_file = -1
        num_attributes_per_time_step = -1
        with io.open(f"{dir_path}{constants.COMMANDS.SLASH_DELIM}{constants.DATASETS.METADATA_FILE_NAME}", 'r',
                     encoding='utf-8') as f:
            metadata_dict = json.loads(f.read())
        if metadata_dict is not None:
            if constants.DATASETS.SCHEMA_PROPERTY in metadata_dict:
                schema = metadata_dict[constants.DATASETS.SCHEMA_PROPERTY]
            else:
                schema = ""
            if constants.DATASETS.NUM_TRACES_PROPERTY in metadata_dict:
                num_traces = metadata_dict[constants.DATASETS.NUM_TRACES_PROPERTY]
            else:
                num_traces = 0
            if constants.DATASETS.NUM_TRACES_PER_FILE_PROPERTY in metadata_dict:
                num_attributes_per_time_step = metadata_dict[constants.DATASETS.NUM_TRACES_PER_FILE_PROPERTY]
            else:
                num_attributes_per_time_step = 0
            if constants.DATASETS.NUM_TRACES_PER_FILE_PROPERTY in metadata_dict:
                num_traces_per_file = metadata_dict[constants.DATASETS.NUM_TRACES_PER_FILE_PROPERTY]
            else:
                num_traces_per_file = 0
            if constants.DATASETS.FILE_FORMAT_PROPERTY in metadata_dict:
                file_format = metadata_dict[constants.DATASETS.FILE_FORMAT_PROPERTY]
            else:
                file_format = "unknown"
            if constants.DATASETS.ADDED_BY_PROPERTY in metadata_dict:
                added_by = metadata_dict[constants.DATASETS.ADDED_BY_PROPERTY]
            else:
                added_by = "unknown"
            if constants.DATASETS.COLUMNS_PROPERTY in metadata_dict:
                columns = metadata_dict[constants.DATASETS.COLUMNS_PROPERTY]
            else:
                columns = ""

        num_files = len([name for name in os.listdir(dir_path) if os.path.isfile(os.path.join(dir_path, name))]) - 1
        size_compressed_gb = round(float(os.path.getsize(zip_file_path)) / 1000000000, 2)
        dir_size_uncompressed_gb = ExportUtil.get_dir_size_gb(dir_path=dir_path)
        return (num_files, dir_size_uncompressed_gb, size_compressed_gb, file_format, num_traces, schema,
                num_traces_per_file, num_attributes_per_time_step, added_by, columns)

    @staticmethod
    def export_emulation_traces_to_disk_csv(
            num_traces_per_file: int, output_dir: str, zip_file_output: str, max_num_traces: int, max_time_steps: int,
            max_nodes: int, max_ports: int, max_vulns: int, null_value: int = -1, added_by: str = "unknown") -> None:
        """
        Exports emulation traces from the metastore to disk

        :param num_traces_per_file: the number of traces per file in the output directory
        :param output_dir: the output directory
        :param zip_file_output: the compressed zip file path
        :param max_num_traces: maximum number of traces
        :param added_by: the person who added the dataset
        :param max_time_steps: the maximum number of time-steps to include in a csv row
        :param max_nodes: the maximum number of nodes to include metrics from
        :param max_ports: the maximum number of ports to include metrics from
        :param max_vulns: the maximum number of vulnerabilities to include metrics from
        :param null_value: the default null value if a metric is missing
        :return: None
        """
        Logger.__call__().get_logger().info(f"Exporting emulation traces to disk (csv), output dir: {output_dir}, "
                                            f"output zip file: {zip_file_output}, "
                                            f"num traces per file: {num_traces_per_file}")
        if not os.path.exists(output_dir):
            os.makedirs(output_dir)
        emulation_traces_ids = MetastoreFacade.list_emulation_traces_ids()
        if len(emulation_traces_ids) > max_num_traces:
            emulation_traces_ids = emulation_traces_ids[0:max_num_traces]
        traces = []
        file_id = 1
        file_name = f"{file_id}.csv"
        last_export = 0
        num_attributes_per_time_step = -1
        columns = None
        for i, id_obj in enumerate(emulation_traces_ids):
            Logger.__call__().get_logger().info(f"Reading trace {i}/{len(emulation_traces_ids)} from the metastore")
            tr = MetastoreFacade.get_emulation_trace(id=id_obj[0])
            if tr is None:
                raise ValueError(f"Could not find an emulation trace with id: {id_obj[0]}")
            tr_values, tr_labels = tr.to_csv_record(max_time_steps=max_time_steps, max_nodes=max_nodes,
                                                    max_ports=max_ports, max_vulns=max_vulns, null_value=null_value)
            if num_attributes_per_time_step == -1:
                num_attributes_per_time_step = tr.num_attributes_per_time_step()
            if columns is None:
                columns = tr_labels

            traces.append(tr_values)
            if i > 0 and ((i % num_traces_per_file == 0) or i == (len(emulation_traces_ids) - 1)):
                Logger.__call__().get_logger().info(f"Exporting traces {last_export + 1}-{i} to file: {file_name}")
                with io.open(f"{output_dir}{constants.COMMANDS.SLASH_DELIM}{file_name}", 'w', encoding='utf-8') as f:
                    writer = csv.writer(f)
                    writer.writerow(columns)
                    for row in traces:
                        writer.writerow(row)
                file_id += 1
                file_name = f"{file_id}.csv"
                traces = []
                last_export = i
        file_format = "csv"
        num_traces = len(emulation_traces_ids)
        with io.open(f"{output_dir}{constants.COMMANDS.SLASH_DELIM}{constants.DATASETS.METADATA_FILE_NAME}", 'w',
                     encoding='utf-8') as f:
            metadata_dict: Dict[str, Any] = {}
            metadata_dict[constants.DATASETS.FILE_FORMAT_PROPERTY] = file_format
            metadata_dict[constants.DATASETS.NUM_TRACES_PROPERTY] = num_traces
            metadata_dict[constants.DATASETS.NUM_ATTRIBUTES_PER_TIME_STEP_PROPERTY] = num_attributes_per_time_step
            metadata_dict[constants.DATASETS.SCHEMA_PROPERTY] = {}
            metadata_dict[constants.DATASETS.NUM_TRACES_PER_FILE_PROPERTY] = num_traces_per_file
            metadata_dict[constants.DATASETS.ADDED_BY_PROPERTY] = added_by
            if columns is not None:
                metadata_dict[constants.DATASETS.COLUMNS_PROPERTY] = ",".join(list(map(lambda x: str(x), columns)))
            else:
                metadata_dict[constants.DATASETS.COLUMNS_PROPERTY] = ""
            f.write(json.dumps(metadata_dict, indent=4, sort_keys=True))
        ExportUtil.zipdir(dir_path=output_dir, file_path=zip_file_output)
        Logger.__call__().get_logger().info(f"Export of emulation traces to disk complete, "
                                            f"output dir:{output_dir}, output zip file: {zip_file_output}")

    @staticmethod
    def export_emulation_statistics_to_disk_json(output_dir: str, zip_file_output: str, statistics_id: int,
                                                 added_by: str = "unknown") -> None:
        """
        Exports emulation statistics from the metastore to disk

        :param output_dir: the output directory
        :param zip_file_output: the compressed zip file path
        :param added_by: the person who added the dataset
        :param statistics_id: the id of the statistics to fetch
        :return: None
        """
        Logger.__call__().get_logger().info(f"Exporting emulation statistics to disk (json), output dir: {output_dir}, "
                                            f"output zip file: {zip_file_output}")
        if not os.path.exists(output_dir):
            os.makedirs(output_dir)
        emulation_statistic = MetastoreFacade.get_emulation_statistic(id=statistics_id)
        if emulation_statistic is None:
            raise ValueError(f"Could not find an emulation statistic with id: {statistics_id}")
        file_name = "statistics.json"
        Logger.__call__().get_logger().info(f"Exporting statistics with id {statistics_id} to file: {file_name}")
        emulation_statistic.compute_descriptive_statistics_and_distributions()
        num_measurements = emulation_statistic.num_measurements
        num_metrics = emulation_statistic.num_metrics
        metrics = ",".join(emulation_statistic.metrics)
        conditions = ",".join(emulation_statistic.conditions)
        num_conditions = emulation_statistic.num_conditions
        statistics_dict = emulation_statistic.to_dict()
        statistics_dict_json = json.dumps(statistics_dict, indent=4, sort_keys=True)
        with io.open(f"{output_dir}{constants.COMMANDS.SLASH_DELIM}{file_name}", 'w', encoding='utf-8') as f:
            f.write(statistics_dict_json)
        file_format = "json"
        with io.open(f"{output_dir}{constants.COMMANDS.SLASH_DELIM}{constants.DATASETS.METADATA_FILE_NAME}", 'w',
                     encoding='utf-8') as f:
            metadata_dict: Dict[str, Any] = {}
            metadata_dict[constants.DATASETS.FILE_FORMAT_PROPERTY] = file_format
            metadata_dict[constants.DATASETS.NUM_MEASUREMENTS_PROPERTY] = num_measurements
            metadata_dict[constants.DATASETS.NUM_CONDITIONS_PROPERTY] = num_conditions
            metadata_dict[constants.DATASETS.NUM_METRICS_PROPERTY] = num_metrics
            metadata_dict[constants.DATASETS.ADDED_BY_PROPERTY] = added_by
            metadata_dict[constants.DATASETS.CONDITIONS_PROPERTY] = conditions
            metadata_dict[constants.DATASETS.METRICS_PROPERTY] = metrics
            f.write(json.dumps(metadata_dict, indent=4, sort_keys=True))
        ExportUtil.zipdir(dir_path=output_dir, file_path=zip_file_output)
        Logger.__call__().get_logger().info(f"Export of emulation statistics to disk complete, "
                                            f"output dir:{output_dir}, output zip file: {zip_file_output}")

    @staticmethod
    def extract_emulation_statistics_dataset_metadata(dir_path: str, zip_file_path: str) \
            -> Tuple[int, float, float, str, str, int, int, str, str, int]:
        """
        Extracts metadata of a traces dataset stored on disk

        :param dir_path: the path to the directory where the traces dataset is stored
        :param zip_file_path: the path to the compressed zipfile of the dataset
        :return: num_files,  dir_size_uncompressed_gb, size_compressed_gb, file_format, added_by, num_measurements,
                 num_metrics, metrics, conditions, num_conditions
        """
        file_format = "unknown"
        added_by = "unknown"
        num_measurements = 0
        num_metrics = 0
        metrics = ""
        conditions = ""
        with io.open(f"{dir_path}{constants.COMMANDS.SLASH_DELIM}{constants.DATASETS.METADATA_FILE_NAME}", 'r',
                     encoding='utf-8') as f:
            metadata_dict = json.loads(f.read())
        if metadata_dict is not None:
            if constants.DATASETS.FILE_FORMAT_PROPERTY in metadata_dict:
                file_format = metadata_dict[constants.DATASETS.FILE_FORMAT_PROPERTY]
            if constants.DATASETS.ADDED_BY_PROPERTY in metadata_dict:
                added_by = metadata_dict[constants.DATASETS.ADDED_BY_PROPERTY]
            if constants.DATASETS.NUM_MEASUREMENTS_PROPERTY in metadata_dict:
                num_measurements = metadata_dict[constants.DATASETS.NUM_MEASUREMENTS_PROPERTY]
            if constants.DATASETS.NUM_METRICS_PROPERTY in metadata_dict:
                num_metrics = metadata_dict[constants.DATASETS.NUM_METRICS_PROPERTY]
            if constants.DATASETS.METRICS_PROPERTY in metadata_dict:
                metrics = metadata_dict[constants.DATASETS.METRICS_PROPERTY]
            if constants.DATASETS.CONDITIONS_PROPERTY in metadata_dict:
                conditions = metadata_dict[constants.DATASETS.CONDITIONS_PROPERTY]
            if constants.DATASETS.NUM_CONDITIONS_PROPERTY in metadata_dict:
                num_conditions = metadata_dict[constants.DATASETS.NUM_CONDITIONS_PROPERTY]

        num_files = len([name for name in os.listdir(dir_path) if os.path.isfile(os.path.join(dir_path, name))]) - 1
        size_compressed_gb = round(float(os.path.getsize(zip_file_path)) / 1000000000, 2)
        dir_size_uncompressed_gb = ExportUtil.get_dir_size_gb(dir_path=dir_path)
        return (num_files, dir_size_uncompressed_gb, size_compressed_gb, file_format, added_by, num_measurements,
                num_metrics, metrics, conditions, num_conditions)



from typing import Optional, List, Set, Dict, Any, Tuple
from csle_common.dao.emulation_config.emulation_env_config import EmulationEnvConfig
from csle_common.dao.emulation_observation.attacker.emulation_attacker_machine_observation_state \
    import EmulationAttackerMachineObservationState
from csle_common.dao.emulation_action.attacker.emulation_attacker_action import EmulationAttackerAction
from csle_common.dao.emulation_action.attacker.emulation_attacker_action_id import EmulationAttackerActionId
from csle_base.json_serializable import JSONSerializable


class EmulationAttackerObservationState(JSONSerializable):
    """
    Represents the attacker's agent's current belief state of the emulation
    """

    def __init__(self, catched_flags: int, agent_reachable: Set[str]):
        """
        Initializes the state

        :param num_flags: the number of flags
        :param catched_flags: the number of catched flags
        :param agent_reachable: whether this node is reachable from the agent
        """
        self.machines: List[EmulationAttackerMachineObservationState] = []
        self.catched_flags = catched_flags
        self.actions_tried: Set[Tuple[int, int, str]] = set()
        self.agent_reachable = agent_reachable

    @staticmethod
    def from_dict(d: Dict[str, Any]) -> "EmulationAttackerObservationState":
        """
        Converts a dict representation to an instance

        :param d: the dict to convert
        :return: the instance
        """
        obj = EmulationAttackerObservationState(catched_flags=d["catched_flags"],
                                                agent_reachable=set(d["agent_reachable"]))
        obj.machines = list(map(lambda x: EmulationAttackerMachineObservationState.from_dict(x), d["machines"]))
        actions_tried = set()
        for i in range(len(d["actions_tried"])):
            actions_tried.add((int(d["actions_tried"][i][0]), int(d["actions_tried"][i][1]),
                               str(d["actions_tried"][i][2])))
        obj.actions_tried = actions_tried
        return obj

    def to_dict(self) -> Dict[str, Any]:
        """
        Converts the object to a dict representation
        
        :return: a dict representation of the object
        """
        d: Dict[str, Any] = {}
        d["machines"] = list(map(lambda x: x.to_dict(), self.machines))
        d["catched_flags"] = self.catched_flags
        d["actions_tried"] = sorted(list(self.actions_tried))
        d["agent_reachable"] = sorted(list(self.agent_reachable))
        return d

    def sort_machines(self) -> None:
        """
        Sorts the machines in the observation

        :return: None
        """
        self.machines = sorted(self.machines, key=lambda x: int(x.ips[0].rsplit(".", 1)[-1]), reverse=False)

    def cleanup(self) -> None:
        """
        Cleanup machine states

        :return: None
        """
        for m in self.machines:
            m.cleanup()

    def get_action_ips(self, a: EmulationAttackerAction, emulation_env_config: EmulationEnvConfig) -> List[str]:
        """
        Returns the ip of the machine that the action targets

        :param a: the action
        :param emulation_env_config: the emulation env config
        :return: the ip of the target
        """
        if a.index == -1 or a.index == len(self.machines):
            return emulation_env_config.topology_config.subnetwork_masks
        elif a.index < len(self.machines):
            return self.machines[a.index].ips
        elif a.index > len(self.machines):
            raise ValueError(f"invalid index: {a.index}, num machines: {len(self.machines)}")
        else:
            return a.ips

    def exploit_tried(self, a: EmulationAttackerAction, m: Optional[EmulationAttackerMachineObservationState]) -> bool:
        """
        Checks if a given exploit have been tried on a given machine or not

        :param a: the exploit action
        :param m: the machine
        :return: true if it has already been tried, otherwise false
        """
        # Known issue with subnet attacks and NMAP: https://github.com/nmap/nmap/issues/1321
        if m is not None:
            if (a.id == EmulationAttackerActionId.SSH_SAME_USER_PASS_DICTIONARY_ALL
                    or a.id == EmulationAttackerActionId.SSH_SAME_USER_PASS_DICTIONARY_HOST):
                return m.ssh_brute_tried

            if (a.id == EmulationAttackerActionId.TELNET_SAME_USER_PASS_DICTIONARY_ALL
                    or a.id == EmulationAttackerActionId.TELNET_SAME_USER_PASS_DICTIONARY_HOST):
                return m.telnet_brute_tried

            if (a.id == EmulationAttackerActionId.FTP_SAME_USER_PASS_DICTIONARY_ALL
                    or a.id == EmulationAttackerActionId.FTP_SAME_USER_PASS_DICTIONARY_HOST):
                return m.ftp_brute_tried

            if (a.id == EmulationAttackerActionId.CASSANDRA_SAME_USER_PASS_DICTIONARY_ALL
                    or a.id == EmulationAttackerActionId.CASSANDRA_SAME_USER_PASS_DICTIONARY_HOST):
                return m.cassandra_brute_tried

            if (a.id == EmulationAttackerActionId.IRC_SAME_USER_PASS_DICTIONARY_ALL
                    or a.id == EmulationAttackerActionId.IRC_SAME_USER_PASS_DICTIONARY_HOST):
                return m.irc_brute_tried

            if (a.id == EmulationAttackerActionId.MONGO_SAME_USER_PASS_DICTIONARY_ALL
                    or a.id == EmulationAttackerActionId.MONGO_SAME_USER_PASS_DICTIONARY_HOST):
                return m.mongo_brute_tried

            if (a.id == EmulationAttackerActionId.MYSQL_SAME_USER_PASS_DICTIONARY_ALL
                    or a.id == EmulationAttackerActionId.MYSQL_SAME_USER_PASS_DICTIONARY_HOST):
                return m.mysql_brute_tried

            if (a.id == EmulationAttackerActionId.SMTP_SAME_USER_PASS_DICTIONARY_ALL or
                    a.id == EmulationAttackerActionId.SMTP_SAME_USER_PASS_DICTIONARY_HOST):
                return m.smtp_brute_tried

            if (a.id == EmulationAttackerActionId.POSTGRES_SAME_USER_PASS_DICTIONARY_ALL or
                    a.id == EmulationAttackerActionId.POSTGRES_SAME_USER_PASS_DICTIONARY_HOST):
                return m.postgres_brute_tried

            if a.id == EmulationAttackerActionId.SAMBACRY_EXPLOIT:
                return m.sambacry_tried

            if a.id == EmulationAttackerActionId.SHELLSHOCK_EXPLOIT:
                return m.shellshock_tried

            if a.id == EmulationAttackerActionId.DVWA_SQL_INJECTION:
                return m.dvwa_sql_injection_tried

            if a.id == EmulationAttackerActionId.CVE_2015_3306_EXPLOIT:
                return m.cve_2015_3306_tried

            if a.id == EmulationAttackerActionId.CVE_2015_1427_EXPLOIT:
                return m.cve_2015_1427_tried

            if a.id == EmulationAttackerActionId.CVE_2016_10033_EXPLOIT:
                return m.cve_2016_10033_tried

            if a.id == EmulationAttackerActionId.CVE_2010_0426_PRIV_ESC:
                return m.cve_2010_0426_tried

            if a.id == EmulationAttackerActionId.CVE_2015_5602_PRIV_ESC:
                return m.cve_2015_5602_tried

            return False
        else:
            exploit_tried = True
            for m2 in self.machines:
                res = self.exploit_tried(a=a, m=m2)
                if not res:
                    exploit_tried = res
                    break
            return exploit_tried

    def exploit_executed(self, machine: EmulationAttackerMachineObservationState) -> bool:
        """
        Check if exploit have been tried on a particular machine

        :param machine: the machine
        :return: true if some exploit have been launched, false otherwise
        """
        if (machine.telnet_brute_tried or machine.ssh_brute_tried or machine.ftp_brute_tried or
                machine.cassandra_brute_tried or machine.irc_brute_tried or machine.mongo_brute_tried or
                machine.mysql_brute_tried or machine.smtp_brute_tried or machine.postgres_brute_tried or
                machine.sambacry_tried or machine.shellshock_tried or machine.dvwa_sql_injection_tried or
                machine.cve_2015_3306_tried or machine.cve_2015_1427_tried or
                machine.cve_2016_10033_tried or machine.cve_2010_0426_tried or machine.cve_2015_5602_tried):
            return True
        return False

    def copy(self) -> "EmulationAttackerObservationState":
        """
        :return: a copy of the state
        """
        c = EmulationAttackerObservationState(catched_flags=self.catched_flags,
                                              agent_reachable=self.agent_reachable)
        c.actions_tried = self.actions_tried
        for m in self.machines:
            c.machines.append(m.copy())
        return c

    def __str__(self) -> str:
        """
        :return: a string representation of the state
        """
        return f"Found flags:{self.catched_flags}," + "\n" + "\n".join(
            [str(i) + ":" + str(self.machines[i]) for i in range(len(self.machines))])

    @staticmethod
    def from_json_file(json_file_path: str) -> "EmulationAttackerObservationState":
        """
        Reads a json file and converts it to a DTO

        :param json_file_path: the json file path
        :return: the converted DTO
        """
        import io
        import json
        with io.open(json_file_path, 'r') as f:
            json_str = f.read()
        return EmulationAttackerObservationState.from_dict(json.loads(json_str))

    def num_attributes(self):
        """
        :return: number of attributes of the DTO
        """
        num_attributes = 3
        if len(self.machines) > 0:
            num_attributes = num_attributes + len(self.machines) * self.machines[0].num_attributes()
        return num_attributes

    @staticmethod
    def schema() -> "EmulationAttackerObservationState":
        """
        :return: get the schema of the DTO
        """
        dto = EmulationAttackerObservationState(catched_flags=0, agent_reachable=set())
        dto.agent_reachable.add("")
        dto.actions_tried.add((-1, -1, ""))
        dto.machines = [EmulationAttackerMachineObservationState.schema()]
        return dto



import subprocess
import sys


class InstallationController:
    """
    Controller managing installation of CSLE
    """

    @staticmethod
    def install_all_emulations() -> None:
        """
        Installs all emulations in the metastore

        :return: None
        """
        cmd = "cd $CSLE_HOME/emulation-system/envs/ && make install"
        p = subprocess.Popen(cmd, stdout=subprocess.PIPE, shell=True)
        while True:
            if p.stdout is None:
                raise ValueError("Cannot read due to None type")
            out = p.stdout.read(1)
            if p.poll() is not None:
                break
            if str(out) != '':
                try:
                    sys.stdout.write(out.decode("utf-8"))
                except Exception:
                    pass
                sys.stdout.flush()

    @staticmethod
    def install_emulation(emulation_name: str) -> None:
        """
        Installs a given emulation in the metastore

        :param emulation_name: the name of the emulation to install

        :return: None
        """
        cmd = f"cd $CSLE_HOME/emulation-system/envs/ && make install_{emulation_name}"
        p = subprocess.Popen(cmd, stdout=subprocess.PIPE, shell=True)
        while True:
            if p.stdout is None:
                raise ValueError("Cannot read due to None type")
            out = p.stdout.read(1)
            if p.poll() is not None:
                break
            if str(out) != '':
                try:
                    sys.stdout.write(out.decode("utf-8"))
                except Exception:
                    pass
                sys.stdout.flush()

    @staticmethod
    def uninstall_emulation(emulation_name: str) -> None:
        """
        Uninstalls a given emulation in the metastore

        :param emulation_name: the name of the emulation to uninstall

        :return: None
        """
        cmd = f"cd $CSLE_HOME/emulation-system/envs/ && make uninstall_{emulation_name}"
        p = subprocess.Popen(cmd, stdout=subprocess.PIPE, shell=True)
        while True:
            if p.stdout is None:
                raise ValueError("stdout is None")
            out = p.stdout.read(1)
            if p.poll() is not None:
                break
            if str(out) != '':
                try:
                    sys.stdout.write(out.decode("utf-8"))
                except Exception:
                    pass
                sys.stdout.flush()

    @staticmethod
    def uninstall_all_emulations() -> None:
        """
        Uninstalls all emulations from the metastoer

        :return: None
        """
        cmd = "cd $CSLE_HOME/emulation-system/envs/ && make uninstall"
        p = subprocess.Popen(cmd, stdout=subprocess.PIPE, shell=True)
        while True:
            if p.stdout is None:
                raise ValueError("Cannot read due to None type")
            out = p.stdout.read(1)
            if p.poll() is not None:
                break
            if str(out) != '':
                try:
                    sys.stdout.write(out.decode("utf-8"))
                except Exception:
                    pass
                sys.stdout.flush()

    @staticmethod
    def install_all_simulations() -> None:
        """
        Installs all simulations in the metastore

        :return: None
        """
        cmd = "cd $CSLE_HOME/simulation-system/envs/ && make install"
        p = subprocess.Popen(cmd, stdout=subprocess.PIPE, shell=True)
        while True:
            if p.stdout is None:
                raise ValueError("stdout is None")
            out = p.stdout.read(1)
            if p.poll() is not None:
                break
            if str(out) != '':
                try:
                    sys.stdout.write(out.decode("utf-8"))
                except Exception:
                    pass
                sys.stdout.flush()

    @staticmethod
    def install_simulation(simulation_name: str) -> None:
        """
        Installs a given simulation in the metastore

        :param simulation_name: the name of the simulation to uninstall
        :return: None
        """
        cmd = f"cd $CSLE_HOME/simulation-system/envs/ && make install_{simulation_name}"
        p = subprocess.Popen(cmd, stdout=subprocess.PIPE, shell=True)
        while True:
            if p.stdout is None:
                raise ValueError("Cannot read due to None type")
            out = p.stdout.read(1)
            if p.poll() is not None:
                break
            if str(out) != '':
                try:
                    sys.stdout.write(out.decode("utf-8"))
                except Exception:
                    pass
                sys.stdout.flush()

    @staticmethod
    def uninstall_simulation(simulation_name: str) -> None:
        """
        Uninstalls a given simulation in the metastore

        :param simulation_name: the name of the simulation to install
        :return: None
        """
        cmd = f"cd $CSLE_HOME/simulation-system/envs/ && make uninstall_{simulation_name}"
        p = subprocess.Popen(cmd, stdout=subprocess.PIPE, shell=True)
        while True:
            if p.stdout is None:
                raise ValueError("Cannot read due to None type")
            out = p.stdout.read(1)
            if p.poll() is not None:
                break
            if str(out) != '':
                try:
                    sys.stdout.write(out.decode("utf-8"))
                except Exception:
                    pass
                sys.stdout.flush()

    @staticmethod
    def uninstall_all_simulations() -> None:
        """
        Uninstalls all simulations from the metastore

        :return: None
        """
        cmd = "cd $CSLE_HOME/simulation-system/envs/ && make uninstall"
        p = subprocess.Popen(cmd, stdout=subprocess.PIPE, shell=True)
        while True:
            if p.stdout is None:
                raise ValueError("Cannot read due to None type")
            out = p.stdout.read(1)
            if p.poll() is not None:
                break
            if str(out) != '':
                try:
                    sys.stdout.write(out.decode("utf-8"))
                except Exception:
                    pass
                sys.stdout.flush()

    @staticmethod
    def install_derived_images() -> None:
        """
        Installs all derived Docker images

        :return: None
        """
        cmd = "cd $CSLE_HOME/emulation-system/derived_images/ && make build"
        p = subprocess.Popen(cmd, stdout=subprocess.PIPE, shell=True)
        while True:
            if p.stdout is None:
                raise ValueError("Cannot read due to None type")
            out = p.stdout.read(1)
            if p.poll() is not None:
                break
            if str(out) != '':
                try:
                    sys.stdout.write(out.decode("utf-8"))
                except Exception:
                    pass
                sys.stdout.flush()

    @staticmethod
    def install_derived_image(image_name: str) -> None:
        """
        Installs a given derived Docker image

        :param image_name: the name of the image to install
        :return: None
        """
        cmd = f"cd $CSLE_HOME/emulation-system/derived_images/ && make {image_name}"
        p = subprocess.Popen(cmd, stdout=subprocess.PIPE, shell=True)
        while True:
            if p.stdout is None:
                raise ValueError("Cannot read due to None type")
            out = p.stdout.read(1)
            if p.poll() is not None:
                break
            if str(out) != '':
                try:
                    sys.stdout.write(out.decode("utf-8"))
                except Exception:
                    pass
                sys.stdout.flush()

    @staticmethod
    def uninstall_derived_images() -> None:
        """
        Uninstalls all derived Docker images

        :return: None
        """
        cmd = "cd $CSLE_HOME/emulation-system/derived_images/ && make rm_image"
        p = subprocess.Popen(cmd, stdout=subprocess.PIPE, shell=True)
        while True:
            if p.stdout is None:
                raise ValueError("Cannot read due to None type")
            out = p.stdout.read(1)
            if p.poll() is not None:
                break
            if str(out) != '':
                try:
                    sys.stdout.write(out.decode("utf-8"))
                except Exception:
                    pass
                sys.stdout.flush()

    @staticmethod
    def uninstall_derived_image(image_name: str) -> None:
        """
        Uninstalls a given derived Docker image

        :param image_name: the name of the image to uninstall
        :return: None
        """
        cmd = f"cd $CSLE_HOME/emulation-system/derived_images/ && make rm_{image_name}"
        p = subprocess.Popen(cmd, stdout=subprocess.PIPE, shell=True)
        while True:
            if p.stdout is None:
                raise ValueError("Cannot read due to None type")
            out = p.stdout.read(1)
            if p.poll() is not None:
                break
            if str(out) != '':
                try:
                    sys.stdout.write(out.decode("utf-8"))
                except Exception:
                    pass
                sys.stdout.flush()

    @staticmethod
    def install_base_images() -> None:
        """
        Installs all base Docker images

        :return: None
        """
        cmd = "cd $CSLE_HOME/emulation-system/base_images/ && make build"
        p = subprocess.Popen(cmd, stdout=subprocess.PIPE, shell=True)
        while True:
            if p.stdout is None:
                raise ValueError("Cannot read due to None type")
            out = p.stdout.read(1)
            if p.poll() is not None:
                break
            if str(out) != '':
                try:
                    sys.stdout.write(out.decode("utf-8"))
                except Exception:
                    pass
                sys.stdout.flush()

    @staticmethod
    def install_base_image(image_name: str) -> None:
        """
        Installs a given base Docker image

        :param image_name: the name of the image to install
        :return: None
        """
        cmd = f"cd $CSLE_HOME/emulation-system/base_images/ && make {image_name}"
        p = subprocess.Popen(cmd, stdout=subprocess.PIPE, shell=True)
        while True:
            if p.stdout is None:
                raise ValueError("Cannot read due to None type")
            out = p.stdout.read(1)
            if p.poll() is not None:
                break
            if str(out) != '':
                try:
                    sys.stdout.write(out.decode("utf-8"))
                except Exception:
                    pass
                sys.stdout.flush()

    @staticmethod
    def uninstall_base_images() -> None:
        """
        Uninstalls all base Docker images

        :return: None
        """
        cmd = "cd $CSLE_HOME/emulation-system/base_images/ && make rm_image"
        p = subprocess.Popen(cmd, stdout=subprocess.PIPE, shell=True)
        while True:
            if p.stdout is None:
                raise ValueError("Cannot read due to None type")
            out = p.stdout.read(1)
            if p.poll() is not None:
                break
            if str(out) != '':
                try:
                    sys.stdout.write(out.decode("utf-8"))
                except Exception:
                    pass
                sys.stdout.flush()

    @staticmethod
    def uninstall_base_image(image_name: str) -> None:
        """
        Uninstalls a given base Docker image

        :param image_name: the name of the image to uninstall
        :return: None
        """
        cmd = f"cd $CSLE_HOME/emulation-system/base_images/ && make rm_{image_name}"
        p = subprocess.Popen(cmd, stdout=subprocess.PIPE, shell=True)
        while True:
            if p.stdout is None:
                raise ValueError("Cannot read due to None type")
            out = p.stdout.read(1)
            if p.poll() is not None:
                break
            if str(out) != '':
                try:
                    sys.stdout.write(out.decode("utf-8"))
                except Exception:
                    pass
                sys.stdout.flush()

    @staticmethod
    def uninstall_metastore() -> None:
        """
        Uninstalls the metastore

        :return: None
        """
        cmd = "cd $CSLE_HOME/metastore/ && make clean"
        p = subprocess.Popen(cmd, stdout=subprocess.PIPE, shell=True)
        while True:
            if p.stdout is None:
                raise ValueError("Cannot read due to None type")
            out = p.stdout.read(1)
            if p.poll() is not None:
                break
            if str(out) != '':
                try:
                    sys.stdout.write(out.decode("utf-8"))
                except Exception:
                    pass
                sys.stdout.flush()

    @staticmethod
    def install_metastore() -> None:
        """
        Installs the metastore

        :return: None
        """
        cmd = "cd $CSLE_HOME/metastore/ && make build"
        p = subprocess.Popen(cmd, stdout=subprocess.PIPE, shell=True)
        while True:
            if p.stdout is None:
                raise ValueError("Cannot read due to None type")
            out = p.stdout.read(1)
            if p.poll() is not None:
                break
            if str(out) != '':
                try:
                    sys.stdout.write(out.decode("utf-8"))
                except Exception:
                    pass
                sys.stdout.flush()



from unittest.mock import patch, MagicMock
from csle_common.util.docker_util import DockerUtil
import csle_common.constants.constants as constants
import json


class TestDockerUtilSuite:
    """
    Test suite for docker util
    """

    @patch("docker.from_env")
    @patch("docker.APIClient")
    @patch("csle_common.util.docker_util.DockerUtil.parse_running_containers")
    @patch("csle_common.util.docker_util.DockerUtil.parse_running_emulation_envs")
    def test_parse_running_emulation_infos(
        self, mock_parse_running_emulation_envs, mock_parse_running_containers, mock_APIClient, mock_from_env
    ) -> None:
        """
        Test method that queries docker to get a list of all running emulation environments

        :param mock_parse_running_emulation_envs: mock_parse_running_emulation_envs
        :param mock_parse_running_containers: mock_parse_running_containers
        :param mock_APIClient: mock_APIClient
        :param mock_from_env: mock_from_env

        :return: None
        """
        mock_client_1 = MagicMock()
        mock_client_2 = MagicMock()
        mock_from_env.return_value = mock_client_1
        mock_APIClient.return_value = mock_client_2
        mock_parsed_containers = [MagicMock()]
        mock_parse_running_containers.return_value = mock_parsed_containers
        mock_parsed_envs = [MagicMock()]
        mock_parse_running_emulation_envs.return_value = mock_parsed_envs
        result = DockerUtil.parse_runnning_emulation_infos()
        mock_from_env.assert_called()
        mock_APIClient.assert_called()
        mock_parse_running_containers.assert_called()
        mock_parse_running_emulation_envs.assert_called()
        assert result == mock_parsed_envs

    @patch("docker.from_env")
    @patch("docker.APIClient")
    @patch("csle_common.util.docker_util.DockerUtil.parse_running_containers")
    def test_get_container_hex_id(self, mock_parse_running_containers, mock_APIClient, mock_from_env) -> None:
        """
        Test method that queries the docker engine for the id of a container with a given name

        :param mock_parse_running_containers: mock_parse_running_containers
        :param mock_APIClient: mock_APIClient
        :param mock_from_env: mock_from_env

        :return: None
        """
        mock_client_1 = MagicMock()
        mock_client_2 = MagicMock()
        mock_from_env.return_value = mock_client_1
        mock_APIClient.return_value = mock_client_2
        container_1 = MagicMock()
        container_1.name = "container_1"
        container_1.id = "1"
        container_2 = MagicMock()
        container_2.name = "container_2"
        container_2.id = "2"
        mock_parse_running_containers.return_value = [container_1, container_2]
        result = DockerUtil.get_container_hex_id("container_1")
        assert result == "1"

    @patch("csle_common.util.docker_util.DockerUtil.parse_containers")
    def test_parse_running_container(self, mock_parse_containers) -> None:
        """
        Test method that queries docker to get a list of all running containers

        :param mock_parse_containers: mock_parse_containers

        :return: None
        """
        mock_client_1 = MagicMock()
        mock_client_2 = MagicMock()
        container_1 = MagicMock()
        container_2 = MagicMock()
        mock_client_1.containers.list.return_value = [container_1, container_2]
        mock_parsed_container_1 = MagicMock()
        mock_parsed_container_2 = MagicMock()
        mock_parse_containers.return_value = [mock_parsed_container_1, mock_parsed_container_2]
        result = DockerUtil.parse_running_containers(client_1=mock_client_1, client_2=mock_client_2)
        mock_client_1.containers.list.assert_called()
        mock_parse_containers.assert_called()
        assert result == [mock_parsed_container_1, mock_parsed_container_2]

    @patch("csle_common.util.docker_util.DockerUtil.parse_containers")
    def test_parse_stopped_container(self, mock_parse_containers) -> None:
        """
        Test method that queries docker to get a list of all stopped csle containers

        :param mock_parse_containers: mock_parse_containers

        :return: None
        """
        mock_client_1 = MagicMock()
        mock_client_2 = MagicMock()
        container_1 = MagicMock()
        container_1.status = constants.DOCKER.CONTAINER_EXIT_STATUS
        container_2 = MagicMock()
        container_2.status = constants.DOCKER.CONTAINER_CREATED_STATUS
        mock_client_1.containers.list.return_value = [container_1, container_2]
        mock_parsed_container_1 = MagicMock()
        mock_parsed_container_2 = MagicMock()
        mock_parse_containers.return_value = [mock_parsed_container_1, mock_parsed_container_2]
        result = DockerUtil.parse_stopped_containers(client_1=mock_client_1, client2=mock_client_2)
        mock_client_1.containers.list.assert_called()
        mock_parse_containers.assert_called()
        assert result == [mock_parsed_container_1, mock_parsed_container_2]

    @patch("csle_common.metastore.metastore_facade.MetastoreFacade.get_emulation_by_name")
    def test_parse_running_emulation_envs(self, mock_get_emulation_by_name) -> None:
        """
        Test method that queries docker to get a list of all active emulation environments

        :param mock_get_emulation_by_name: mock_get_emulation_by_name

        :return: None
        """
        container_1 = MagicMock()
        container_1.emulation = "emulation1"
        container_1.ip = "192.168.1.2"
        container_1.ip_prefix_len = 24
        container_1.level = "level1"

        container_2 = MagicMock()
        container_2.emulation = "emulation1"
        container_2.ip = "192.168.1.3"
        container_2.ip_prefix_len = 24
        container_2.level = "level1"

        containers = [container_1, container_2]
        emulations = ["emulation1"]

        mock_em_record = MagicMock()
        mock_get_emulation_by_name.return_value = mock_em_record

        result = DockerUtil.parse_running_emulation_envs(emulations=emulations, containers=containers)
        mock_get_emulation_by_name.assert_called()
        assert result

    @patch("docker.APIClient.inspect_container")
    def test_parse_containers(self, mock_inspect_container) -> None:
        """
        Test method that queries docker to get a list of running or stopped csle containers

        :param mock_inspect_container: mock_inspect_container

        :return: None
        """
        mock_client2 = MagicMock()
        mock_inspect_info = {
            constants.DOCKER.NETWORK_SETTINGS: {
                constants.DOCKER.NETWORKS: {
                    "net": {
                        constants.DOCKER.IP_ADDRESS_INFO: "192.168.1.2",
                        constants.DOCKER.NETWORK_ID_INFO: "network_1",
                        constants.DOCKER.GATEWAY_INFO: "192.168.1.1",
                        constants.DOCKER.MAC_ADDRESS_INFO: "02:42:ac:11:00:02",
                        constants.DOCKER.IP_PREFIX_LEN_INFO: 24,
                    }
                }
            },
            constants.DOCKER.CREATED_INFO: "2021-01-01T00:00:00Z",
            constants.DOCKER.CONFIG: {constants.DOCKER.HOSTNAME_INFO: "container_1", constants.DOCKER.IMAGE: "image_1"},
        }
        mock_inspect_container.return_value = mock_inspect_info

        container_1 = MagicMock()
        container_1.name = "csle_container_1-level1-1"
        container_1.status = "running"
        container_1.short_id = "1"
        container_1.image.short_id = "image1"
        container_1.image.tags = ["latest"]
        container_1.id = "container_id_1"
        container_1.labels = {
            constants.DOCKER.CFG: "/path/to/config",
            constants.DOCKER.CONTAINER_CONFIG_DIR: "/path/to/dir",
            constants.DOCKER.EMULATION: "emulation_1",
            constants.DOCKER.KAFKA_CONFIG: "kafka_config_1",
        }
        containers = [container_1]
        result = DockerUtil.parse_containers(containers=containers, client2=mock_client2)
        # mock_inspect_container.assert_called_once_with("container_id_1")
        assert result

    @patch("os.popen")
    def test_get_docker_gw_bridge_ip(self, mock_popen) -> None:
        """
        Test method that gets the docker gw bridge ip of a container

        :param mock_popen: mock_popen

        :return: None
        """
        mock_json_output = json.dumps(
            [{constants.DOCKER.CONTAINERS_KEY: {"container_id_1": {constants.DOCKER.IPV4_KEY: "192.168.1.2/24"}}}]
        )
        mock_stream = MagicMock()
        mock_stream.read.return_value = mock_json_output
        mock_popen.return_value = mock_stream
        result = DockerUtil.get_docker_gw_bridge_ip("container_id_1")
        assert result == "192.168.1.2"



from typing import List, Dict, Any
from csle_common.dao.training.agent_type import AgentType
from csle_common.dao.training.player_type import PlayerType
from csle_common.dao.training.hparam import HParam
from csle_base.json_serializable import JSONSerializable


class ExperimentConfig(JSONSerializable):
    """
    DTO representing the configuration of an experiment
    """

    def __init__(self, output_dir: str, title: str, random_seeds: List[int], agent_type: AgentType,
                 hparams: Dict[str, HParam], log_every: int, player_type: PlayerType, player_idx: int,
                 br_log_every: int = 10):
        """
        Initializes the DTO

        :param output_dir: the directory to save logs etc of the experiment
        :param title: the title of the experiment
        :param random_seeds: the random seeds to use in the experiment
        :param agent_type: the type of agent for the experiment
        :param hparams: the hyperparameters
        :param log_every: frequency of logging
        :param player_type: the type of player (e.g. defender or attacker)
        :param player_idx: the player index
        :param br_log_every: frequency of logging when training best response strategies (for game-theoretic algorithms)
        """
        self.output_dir = output_dir
        self.title = title
        self.random_seeds = random_seeds
        self.agent_type = agent_type
        self.hparams = hparams
        self.log_every = log_every
        self.player_type = player_type
        self.player_idx = player_idx
        self.br_log_every = br_log_every

    @staticmethod
    def from_dict(d: Dict[str, Any]) -> "ExperimentConfig":
        """
        Converts a dict representation to an instance

        :param d: the dict to convert
        :return: the created instance
        """

        h_d = {}
        for k, v in d["hparams"].items():
            h_d[k] = HParam.from_dict(v)
        obj = ExperimentConfig(
            output_dir=d["output_dir"], title=d["title"], random_seeds=d["random_seeds"],
            agent_type=d["agent_type"], hparams=h_d, log_every=d["log_every"], player_type=d["player_type"],
            player_idx=d["player_idx"], br_log_every=d["br_log_every"]
        )
        return obj

    def to_dict(self) -> Dict[str, Any]:
        """
        Converts the object to a dict representation
        
        :return: a dict representation of the object
        """
        d: Dict[str, Any] = {}
        d["output_dir"] = self.output_dir
        d["title"] = self.title
        d["random_seeds"] = self.random_seeds
        d["agent_type"] = self.agent_type
        d_h = {}
        for k, v in self.hparams.items():
            d_h[k] = v.to_dict()
        d["hparams"] = d_h
        d["log_every"] = self.log_every
        d["player_type"] = self.player_type
        d["player_idx"] = self.player_idx
        d["br_log_every"] = self.br_log_every
        return d

    def __str__(self) -> str:
        """
        :return: a string representation of the object
        """
        return f"output_dir:{self.output_dir}, title:{self.title}, random_seeds:{self.random_seeds}, " \
               f"agent_type: {self.agent_type}, hparams: {self.hparams}, log_every: {self.log_every}, " \
               f"player_type: {self.player_type}, player_idx: {self.player_idx}, br_log_every: {self.br_log_every}"

    @staticmethod
    def from_json_file(json_file_path: str) -> "ExperimentConfig":
        """
        Reads a json file and converts it to a DTO

        :param json_file_path: the json file path
        :return: the converted DTO
        """
        import io
        import json
        with io.open(json_file_path, 'r') as f:
            json_str = f.read()
        return ExperimentConfig.from_dict(json.loads(json_str))



from typing import List, Dict, Any, Optional, Union
import socket
import paramiko
from confluent_kafka import Producer
import csle_common.constants.constants as constants
from csle_common.dao.emulation_config.containers_config import ContainersConfig
from csle_common.dao.emulation_config.node_container_config import NodeContainerConfig
from csle_common.dao.emulation_config.users_config import UsersConfig
from csle_common.dao.emulation_config.flags_config import FlagsConfig
from csle_common.dao.emulation_config.vulnerabilities_config import VulnerabilitiesConfig
from csle_common.dao.emulation_config.topology_config import TopologyConfig
from csle_common.dao.emulation_config.traffic_config import TrafficConfig
from csle_common.dao.emulation_config.resources_config import ResourcesConfig
from csle_common.dao.emulation_config.kafka_config import KafkaConfig
from csle_common.dao.emulation_config.services_config import ServicesConfig
from csle_common.dao.emulation_config.ovs_config import OVSConfig
from csle_common.dao.emulation_config.sdn_controller_config import SDNControllerConfig
from csle_common.dao.emulation_action.attacker.emulation_attacker_action import EmulationAttackerAction
from csle_common.dao.emulation_config.host_manager_config import HostManagerConfig
from csle_common.dao.emulation_config.snort_ids_manager_config import SnortIDSManagerConfig
from csle_common.dao.emulation_config.ossec_ids_manager_config import OSSECIDSManagerConfig
from csle_common.dao.emulation_config.docker_stats_manager_config import DockerStatsManagerConfig
from csle_common.dao.emulation_config.beats_config import BeatsConfig
from csle_common.dao.emulation_config.elk_config import ElkConfig
from csle_common.dao.emulation_config.five_g_config import FiveGConfig
from csle_common.dao.emulation_config.container_network import ContainerNetwork
from csle_common.util.ssh_util import SSHUtil
from csle_common.logging.log import Logger
import csle_collector.constants.constants as collector_constants
from csle_base.json_serializable import JSONSerializable


class EmulationEnvConfig(JSONSerializable):
    """
    Class representing the configuration of an emulation
    """

    def __init__(self, name: str, containers_config: ContainersConfig, users_config: UsersConfig,
                 flags_config: FlagsConfig,
                 vuln_config: VulnerabilitiesConfig, topology_config: TopologyConfig, traffic_config: TrafficConfig,
                 resources_config: ResourcesConfig, kafka_config: KafkaConfig, services_config: ServicesConfig,
                 descr: str, static_attacker_sequences: Dict[str, List[EmulationAttackerAction]],
                 ovs_config: OVSConfig, sdn_controller_config: Optional[SDNControllerConfig],
                 host_manager_config: HostManagerConfig, snort_ids_manager_config: SnortIDSManagerConfig,
                 ossec_ids_manager_config: OSSECIDSManagerConfig,
                 docker_stats_manager_config: DockerStatsManagerConfig, elk_config: ElkConfig,
                 five_g_config: FiveGConfig, beats_config: BeatsConfig,
                 level: int, version: str, execution_id: int,
                 csle_collector_version: str = collector_constants.LATEST_VERSION,
                 csle_ryu_version: str = collector_constants.LATEST_VERSION):
        """
        Initializes the object

        :param name: the name of the emulation
        :param containers_config: the containers configuration
        :param users_config: the users configuration
        :param flags_config: the flags configuration
        :param vuln_config: the vulnerabilities configuration
        :param topology_config: the topology configuration
        :param traffic_config: the traffic configuration
        :param resources_config: the resources configuration
        :param services_config: the services configuration
        :param descr: a description of the environment configuration
        :param static_attacker_sequences: dict with static attacker sequences
        :param ovs_config: the OVS config
        :param sdn_controller_config: the SDN controller config
        :param host_manager_config: the host manager config
        :param snort_ids_manager_config: the Snort IDS manager config
        :param ossec_ids_manager_config: the OSSEC IDS manager config
        :param docker_stats_manager_config: the Docker stats manager config
        :param beats_config: the beats config
        :param elk_config: the ELK config
        :param five_g_config: the 5G config
        :param level: the level of the emulation
        :param version: the version of the emulation
        :param execution_id: the execution id of the emulation
        :param csle_collector_version: the version of the CSLE collector library
        :param csle_ryu_version: the version of the CSLE Ryu library
        """
        self.name = name
        self.descr = descr
        self.containers_config = containers_config
        self.users_config = users_config
        self.flags_config = flags_config
        self.vuln_config = vuln_config
        self.topology_config = topology_config
        self.traffic_config = traffic_config
        self.resources_config = resources_config
        self.kafka_config = kafka_config
        self.services_config = services_config
        self.connections: Dict[str, paramiko.SSHClient] = {}
        self.producer: Union[None, Any] = None
        self.hostname = socket.gethostname()
        self.port_forward_port = 1900
        self.running = False
        self.image = None
        self.id = -1
        self.static_attacker_sequences = static_attacker_sequences
        self.ovs_config = ovs_config
        self.sdn_controller_config = sdn_controller_config
        self.level = level
        self.execution_id = execution_id
        self.version = version
        self.host_manager_config = host_manager_config
        self.snort_ids_manager_config = snort_ids_manager_config
        self.ossec_ids_manager_config = ossec_ids_manager_config
        self.docker_stats_manager_config = docker_stats_manager_config
        self.elk_config = elk_config
        self.beats_config = beats_config
        self.five_g_config = five_g_config
        self.csle_collector_version = csle_collector_version
        self.csle_ryu_version = csle_ryu_version

    @staticmethod
    def from_dict(d: Dict[str, Any]) -> "EmulationEnvConfig":
        """
        Converts a dict representation into an instance

        :param d: the dict to convert
        :return: the created instance
        """
        static_attacker_sequences = {}
        for k, v in d["static_attacker_sequences"].items():
            static_attacker_sequences[k] = list(map(lambda x: EmulationAttackerAction.from_dict(x), v))
        sdn_controller_config = None
        if d["sdn_controller_config"] is not None:
            sdn_controller_config = SDNControllerConfig.from_dict(d["sdn_controller_config"])
        obj = EmulationEnvConfig(
            name=d["name"], containers_config=ContainersConfig.from_dict(d["containers_config"]),
            users_config=UsersConfig.from_dict(d["users_config"]),
            flags_config=FlagsConfig.from_dict(d["flags_config"]),
            vuln_config=VulnerabilitiesConfig.from_dict(d["vuln_config"]),
            topology_config=TopologyConfig.from_dict(d["topology_config"]),
            traffic_config=TrafficConfig.from_dict(d["traffic_config"]),
            resources_config=ResourcesConfig.from_dict(d["resources_config"]),
            kafka_config=KafkaConfig.from_dict(d["kafka_config"]),
            services_config=ServicesConfig.from_dict(d["services_config"]),
            descr=d["descr"], static_attacker_sequences=static_attacker_sequences,
            ovs_config=OVSConfig.from_dict(d["ovs_config"]),
            sdn_controller_config=sdn_controller_config,
            level=d["level"], execution_id=d["execution_id"], version=d["version"],
            host_manager_config=HostManagerConfig.from_dict(d["host_manager_config"]),
            ossec_ids_manager_config=OSSECIDSManagerConfig.from_dict(d["ossec_ids_manager_config"]),
            snort_ids_manager_config=SnortIDSManagerConfig.from_dict(d["snort_ids_manager_config"]),
            docker_stats_manager_config=DockerStatsManagerConfig.from_dict(d["docker_stats_manager_config"]),
            elk_config=ElkConfig.from_dict(d["elk_config"]), csle_collector_version=d["csle_collector_version"],
            beats_config=BeatsConfig.from_dict(d["beats_config"]), csle_ryu_version=d["csle_ryu_version"],
            five_g_config=FiveGConfig.from_dict(d["five_g_config"]),
        )
        obj.running = d["running"]
        obj.image = d["image"]
        obj.id = d["id"]
        return obj

    def to_dict(self) -> Dict[str, Any]:
        """
        Converts the object to a dict representation
        
        :return: a dict representation of the object
        """
        d: Dict[str, Any] = {}
        d["name"] = self.name
        d["containers_config"] = self.containers_config.to_dict()
        d["users_config"] = self.users_config.to_dict()
        d["flags_config"] = self.flags_config.to_dict()
        d["vuln_config"] = self.vuln_config.to_dict()
        d["topology_config"] = self.topology_config.to_dict()
        d["traffic_config"] = self.traffic_config.to_dict()
        d["resources_config"] = self.resources_config.to_dict()
        d["kafka_config"] = self.kafka_config.to_dict()
        d["services_config"] = self.services_config.to_dict()
        d["hostname"] = self.hostname
        d["running"] = self.running
        d["image"] = self.image
        d["descr"] = self.descr
        d["id"] = self.id
        d["version"] = self.version
        d["level"] = self.level
        d["execution_id"] = self.execution_id
        d["ovs_config"] = self.ovs_config.to_dict()
        if self.sdn_controller_config is not None:
            d["sdn_controller_config"] = self.sdn_controller_config.to_dict()
        else:
            d["sdn_controller_config"] = None
        d2 = {}
        for k, v in self.static_attacker_sequences.items():
            d2[k] = list(map(lambda x: x.to_dict(), v))
        d["static_attacker_sequences"] = d2
        d["host_manager_config"] = self.host_manager_config.to_dict()
        d["snort_ids_manager_config"] = self.snort_ids_manager_config.to_dict()
        d["ossec_ids_manager_config"] = self.ossec_ids_manager_config.to_dict()
        d["docker_stats_manager_config"] = self.docker_stats_manager_config.to_dict()
        d["elk_config"] = self.elk_config.to_dict()
        d["five_g_config"] = self.five_g_config.to_dict()
        d["csle_collector_version"] = self.csle_collector_version
        d["beats_config"] = self.beats_config.to_dict()
        d["csle_ryu_version"] = self.csle_ryu_version
        return d

    def connect(self, ip: str = "", username: str = "", pw: str = "",
                create_producer: bool = False) -> paramiko.SSHClient:
        """
        Connects to the agent's host with SSH, either directly or through a jumphost

        :param ip: the ip to connect to
        :param username: the username to connect with
        :param pw: the password to connect with
        :param create_producer: whether the producer should be created if it not already created

        :return: the created conn
        """
        if ip in self.connections:
            old_conn = self.connections[ip]
            old_conn.close()
        if ip in self.connections:
            self.connections.pop(ip)
        Logger.__call__().get_logger().info(f"Connecting to host: {ip}")
        conn = paramiko.SSHClient()
        conn.set_missing_host_key_policy(paramiko.AutoAddPolicy())
        conn.connect(ip, username=username, password=pw)
        transport = conn.get_transport()
        if transport is not None:
            transport.set_keepalive(5)
        else:
            raise Exception("Connection Failed")
        self.connections[ip] = conn
        if self.producer is None and create_producer:
            self.create_producer()

        Logger.__call__().get_logger().info("Connected successfully")
        return conn

    def get_connection(self, ip: str) -> paramiko.SSHClient:
        """
        Gets a connection to a given IP address

        :param ip: the ip address to get the connection for
        :return: the connection
        """
        if ip in self.connections and EmulationEnvConfig.check_if_ssh_connection_is_alive(self.connections[ip]):
            return self.connections[ip]
        else:
            raise ConnectionError(f"Connection to ip:{ip} is not active")

    def get_hacker_connection(self) -> paramiko.SSHClient:
        """
        Gets an SSH connection to the hacker agent, creates one if it does not exist

        :return: SSH connection to the hacker
        """
        container = self.containers_config.get_agent_container()
        if container is None:
            raise ValueError("No agent_container obtained")
        else:
            hacker_ip = container.docker_gw_bridge_ip
        connected = False
        if hacker_ip in self.connections and self.connections[hacker_ip] is not None:
            transport = self.connections[hacker_ip].get_transport()
            if transport is not None and transport.is_active():
                try:
                    SSHUtil.execute_ssh_cmds(cmds=["ls > /dev/null"], conn=self.connections[hacker_ip])
                    connected = True
                except Exception:
                    pass
        if not connected:
            self.connect(ip=hacker_ip, username=constants.AGENT.USER, pw=constants.AGENT.PW, create_producer=False)
        return self.connections[hacker_ip]

    def cleanup(self) -> None:
        """
        Cleans up old connections

        :return: None
        """
        for ip, conn in self.connections.items():
            conn.close()
        self.connections = {}

    def create_producer(self) -> None:
        """
        Creates a Kafka producer

        :return: None
        """
        conf = {
            collector_constants.KAFKA.BOOTSTRAP_SERVERS_PROPERTY:
                f"{self.kafka_config.container.get_ips()[0]}:{self.kafka_config.kafka_port}",
            collector_constants.KAFKA.CLIENT_ID_PROPERTY: self.hostname}
        self.producer = Producer(**conf)

    def close_all_connections(self) -> None:
        """
        Closes the emulation connection
        :return: None
        """
        for k, v in self.connections.items():
            v.close()
        self.connections = {}

    @staticmethod
    def check_if_ssh_connection_is_alive(conn: paramiko.SSHClient) -> bool:
        """
        Utility function to check whether a SSH connection is alive or not
        :param conn: the connection to check
        :return: true or false
        """
        alive = False
        transport = conn.get_transport()
        if transport is not None:
            alive = transport.is_active()
        else:
            raise ValueError("Could not obtain transport")
        return alive

    def get_port_forward_port(self) -> int:
        """
        :return: the next port to use for forwarding
        """
        self.port_forward_port += 1
        return self.port_forward_port

    def ids(self) -> bool:
        """
        Check if the configuration includes an IDS

        :return: True if it includes an IDS, otherwise False
        """
        for c in self.containers_config.containers:
            if c.name in constants.CONTAINER_IMAGES.SNORT_IDS_IMAGES:
                return True
        return False

    def get_container_from_ip(self, ip: str) -> Union[NodeContainerConfig, None]:
        """
        Utility function for getting a container with a specific IP

        :param ip: the ip of the container
        :return: the container with the given ip or None
        """
        if ip in self.kafka_config.container.get_ips():
            return self.kafka_config.container
        if ip in self.elk_config.container.get_ips():
            return self.elk_config.container
        if self.sdn_controller_config is not None and ip in self.sdn_controller_config.container.get_ips():
            return self.sdn_controller_config.container
        for c in self.containers_config.containers:
            if ip in c.get_ips():
                return c
        return None

    def __str__(self) -> str:
        """
        :return:  a string representation of the object
        """
        return f"name: {self.name}, containers_config: {self.containers_config}, users_config: {self.users_config}, " \
               f"flags_config: {self.flags_config}, vuln_config: {self.vuln_config}, " \
               f"topology_config: {self.topology_config}, traffic_config: {self.traffic_config}, " \
               f"resources_config: {self.resources_config}, kafka_config:{self.kafka_config}, " \
               f"services_config: {self.services_config}, hostname:{self.hostname}, running: {self.running}, " \
               f"descr: {self.descr}, id:{self.id}, static_attacker_sequences: {self.static_attacker_sequences}," \
               f"ovs_config: {self.ovs_config}, sdn_controller_config: {self.sdn_controller_config}," \
               f" host_manager_config: {self.host_manager_config}, " \
               f"snort_ids_manager_config: {self.snort_ids_manager_config}, " \
               f"ossec_ids_manager_config: {self.ossec_ids_manager_config}, " \
               f"docker_stats_manager_config: {self.docker_stats_manager_config}, elk_config: {self.elk_config}," \
               f" csle_collector_version: {self.csle_collector_version}, beats_config: {self.beats_config}," \
               f" csle_ryu_version: {self.csle_ryu_version}, five_g_config: {self.five_g_config},"

    def get_all_ips(self) -> List[str]:
        """
        :return: a list of all ip addresses in the emulation
        """
        ips = set()
        for c in self.containers_config.containers:
            for ip_net in c.ips_and_networks:
                ip, _ = ip_net
                ips.add(ip)
        return list(ips)

    @staticmethod
    def from_json_file(json_file_path: str) -> "EmulationEnvConfig":
        """
        Reads a json file and converts it to a DTO

        :param json_file_path: the json file path
        :return: the converted DTO
        """
        import io
        import json
        with io.open(json_file_path, 'r') as f:
            json_str = f.read()
        return EmulationEnvConfig.from_dict(json.loads(json_str))

    def copy(self) -> "EmulationEnvConfig":
        """
        :return: a copy of the DTO
        """
        return EmulationEnvConfig.from_dict(self.to_dict())

    def create_execution_config(self, ip_first_octet: int, physical_servers: List[str]) -> "EmulationEnvConfig":
        """
        Creates an execution config from the base config

        :param ip_first_octet:  the id of the execution
        :param physical_servers: the list of physical servers
        :return: the created execution config
        """
        config = self.copy()
        config.execution_id = ip_first_octet
        config.containers_config = config.containers_config.create_execution_config(
            ip_first_octet=ip_first_octet, physical_servers=physical_servers)
        config.users_config = config.users_config.create_execution_config(ip_first_octet=ip_first_octet)
        config.flags_config = config.flags_config.create_execution_config(ip_first_octet=ip_first_octet)
        config.vuln_config = config.vuln_config.create_execution_config(ip_first_octet=ip_first_octet)
        config.topology_config = config.topology_config.create_execution_config(ip_first_octet=ip_first_octet)
        config.traffic_config = config.traffic_config.create_execution_config(ip_first_octet=ip_first_octet)
        config.resources_config = config.resources_config.create_execution_config(ip_first_octet=ip_first_octet)
        config.kafka_config = config.kafka_config.create_execution_config(ip_first_octet=ip_first_octet,
                                                                          physical_servers=physical_servers)
        config.services_config = config.services_config.create_execution_config(ip_first_octet=ip_first_octet)
        config.ovs_config = config.ovs_config.create_execution_config(ip_first_octet=ip_first_octet,
                                                                      physical_servers=physical_servers)
        config.host_manager_config = config.host_manager_config.create_execution_config(ip_first_octet=ip_first_octet)
        config.snort_ids_manager_config = config.snort_ids_manager_config.create_execution_config(
            ip_first_octet=ip_first_octet)
        config.ossec_ids_manager_config = config.ossec_ids_manager_config.create_execution_config(
            ip_first_octet=ip_first_octet)
        config.docker_stats_manager_config = config.docker_stats_manager_config.create_execution_config(
            ip_first_octet=ip_first_octet)
        config.elk_config = config.elk_config.create_execution_config(
            ip_first_octet=ip_first_octet, physical_servers=physical_servers)
        config.beats_config = config.beats_config.create_execution_config(ip_first_octet=ip_first_octet)
        config.five_g_config = config.five_g_config.create_execution_config(ip_first_octet=ip_first_octet)
        if config.sdn_controller_config is not None:
            config.sdn_controller_config = config.sdn_controller_config.create_execution_config(
                ip_first_octet=ip_first_octet, physical_servers=physical_servers)
        static_attacker_sequences = {}
        for k, v in config.static_attacker_sequences.items():
            static_attacker_sequences[k] = list(map(lambda x: x.create_execution_config(ip_first_octet=ip_first_octet),
                                                    config.static_attacker_sequences[k]))
        config.static_attacker_sequences = static_attacker_sequences
        return config

    def get_network_by_name(self, net_name: str) -> Optional[ContainerNetwork]:
        """
        Gets the network by name from the config

        :param net_name: the name of the network
        :return: the network with the given name
        """
        for c in self.containers_config.networks:
            if c.name == net_name:
                return c
            
        return None



from csle_common.dao.training.alpha_vectors_policy import AlphaVectorsPolicy
from csle_common.dao.training.dqn_policy import DQNPolicy
from csle_common.dao.training.experiment_config import ExperimentConfig
from csle_common.dao.training.experiment_execution import ExperimentExecution
from csle_common.dao.training.experiment_result import ExperimentResult
from csle_common.dao.training.fnn_with_softmax_policy import FNNWithSoftmaxPolicy
from csle_common.dao.training.linear_tabular_policy import LinearTabularPolicy
from csle_common.dao.training.linear_threshold_stopping_policy import LinearThresholdStoppingPolicy
from csle_common.dao.training.tabular_policy import TabularPolicy
from csle_common.dao.training.mixed_linear_tabular import MixedLinearTabularPolicy
from csle_common.dao.training.mixed_ppo_policy import MixedPPOPolicy
from csle_common.dao.training.ppo_policy import PPOPolicy
from csle_common.dao.training.random_policy import RandomPolicy
from csle_common.dao.training.vector_policy import VectorPolicy
from csle_common.dao.training.hparam import HParam
from csle_common.dao.training.mixed_multi_threshold_stopping_policy import MixedMultiThresholdStoppingPolicy
from csle_common.dao.training.multi_threshold_stopping_policy import MultiThresholdStoppingPolicy


class TestTrainingDaoSuite:
    """
    Test suite for training data access objects (DAOs)
    """

    def test_alpha_vectors_policy(self, example_alpha_vectors_policy: AlphaVectorsPolicy) -> None:
        """
        Tests creation and dict conversion of the AlphaVectorsPolicy DAO

        :param example_alpha_vectors_policy: an example AlphaVectorsPolicy
        :return: None
        """
        assert isinstance(example_alpha_vectors_policy.to_dict(), dict)
        assert isinstance(AlphaVectorsPolicy.from_dict(example_alpha_vectors_policy.to_dict()), AlphaVectorsPolicy)
        assert (AlphaVectorsPolicy.from_dict(example_alpha_vectors_policy.to_dict()).to_dict() ==
                example_alpha_vectors_policy.to_dict())
        assert AlphaVectorsPolicy.from_dict(example_alpha_vectors_policy.to_dict()) == example_alpha_vectors_policy

    def test_dqn_policy(self, example_dqn_policy: DQNPolicy) -> None:
        """
        Tests creation and dict conversion of the DQNPolicy DAO

        :param example_dqn_policy: an example DQNPolicy
        :return: None
        """
        assert isinstance(example_dqn_policy.to_dict(), dict)
        assert isinstance(DQNPolicy.from_dict(example_dqn_policy.to_dict()), DQNPolicy)
        assert DQNPolicy.from_dict(example_dqn_policy.to_dict()).to_dict() == example_dqn_policy.to_dict()
        assert DQNPolicy.from_dict(example_dqn_policy.to_dict()) == example_dqn_policy

    def test_experiment_config(self, example_experiment_config: ExperimentConfig) -> None:
        """
        Tests creation and dict conversion of the ExperimentConfig DAO

        :param example_experiment_config: an example ExperimentConfig
        :return: None
        """
        assert isinstance(example_experiment_config.to_dict(), dict)
        assert isinstance(ExperimentConfig.from_dict(example_experiment_config.to_dict()), ExperimentConfig)
        assert ExperimentConfig.from_dict(example_experiment_config.to_dict()).to_dict() == \
               example_experiment_config.to_dict()
        assert ExperimentConfig.from_dict(example_experiment_config.to_dict()) == example_experiment_config

    def test_experiment_execution(self, example_experiment_execution: ExperimentExecution) -> None:
        """
        Tests creation and dict conversion of the ExperimentExecution DAO

        :param example_experiment_execution: an example ExperimentExecution
        :return: None
        """
        assert isinstance(example_experiment_execution.to_dict(), dict)
        assert isinstance(ExperimentExecution.from_dict(example_experiment_execution.to_dict()), ExperimentExecution)
        assert (ExperimentExecution.from_dict(example_experiment_execution.to_dict()).to_dict() ==
                example_experiment_execution.to_dict())
        assert ExperimentExecution.from_dict(example_experiment_execution.to_dict()) == example_experiment_execution

    def test_experiment_result(self, example_experiment_result: ExperimentResult) -> None:
        """
        Tests creation and dict conversion of the ExperimentResult DAO

        :param example_experiment_result: an example ExperimentResult
        :return: None
        """
        assert isinstance(example_experiment_result.to_dict(), dict)
        assert isinstance(ExperimentResult.from_dict(example_experiment_result.to_dict()), ExperimentResult)
        assert ExperimentResult.from_dict(example_experiment_result.to_dict()).to_dict() \
               == example_experiment_result.to_dict()
        assert ExperimentResult.from_dict(example_experiment_result.to_dict()) == example_experiment_result

    def test_fnn_with_softmax_policy(self, example_fnn_with_softmax_policy: FNNWithSoftmaxPolicy) -> None:
        """
        Tests creation and dict conversion of the FNNWithSoftmaxPolicy DAO

        :param example_fnn_with_softmax_policy: an example FNNWithSoftmaxPolicy
        :return: None
        """
        assert isinstance(example_fnn_with_softmax_policy.to_dict(), dict)
        assert isinstance(FNNWithSoftmaxPolicy.from_dict(example_fnn_with_softmax_policy.to_dict()),
                          FNNWithSoftmaxPolicy)
        assert (FNNWithSoftmaxPolicy.from_dict(example_fnn_with_softmax_policy.to_dict()).to_dict() ==
                example_fnn_with_softmax_policy.to_dict())
        assert FNNWithSoftmaxPolicy.from_dict(example_fnn_with_softmax_policy.to_dict()) == \
               example_fnn_with_softmax_policy

    def test_hparam(self, example_hparam: HParam) -> None:
        """
        Tests creation and dict conversion of the HParam DAO

        :param example_hparam: an example y
        :return: None
        """
        assert isinstance(example_hparam.to_dict(), dict)
        assert isinstance(HParam.from_dict(example_hparam.to_dict()), HParam)
        assert HParam.from_dict(example_hparam.to_dict()).to_dict() == example_hparam.to_dict()
        assert HParam.from_dict(example_hparam.to_dict()) == example_hparam

    def test_tabular_policy(self, example_tabular_policy: TabularPolicy) -> None:
        """
        Tests creation and dict conversion of the TabularPolicy DAO

        :param example_tabular_policy: an example TabularPolicy
        :return: None
        """
        assert isinstance(example_tabular_policy.to_dict(), dict)
        assert isinstance(TabularPolicy.from_dict(example_tabular_policy.to_dict()), TabularPolicy)
        assert TabularPolicy.from_dict(example_tabular_policy.to_dict()).to_dict() == example_tabular_policy.to_dict()
        assert TabularPolicy.from_dict(example_tabular_policy.to_dict()) == example_tabular_policy

    def test_linear_threshold_stopping_policy(
            self, example_linear_threshold_stopping_policy: LinearThresholdStoppingPolicy) -> None:
        """
        Tests creation and dict conversion of the LinearThresholdStoppingPolicy DAO

        :param example_linear_threshold_stopping_policy: an example LinearThresholdStoppingPolicy
        :return: None
        """
        assert isinstance(example_linear_threshold_stopping_policy.to_dict(), dict)
        assert isinstance(LinearThresholdStoppingPolicy.from_dict(example_linear_threshold_stopping_policy.to_dict()),
                          LinearThresholdStoppingPolicy)
        assert (LinearThresholdStoppingPolicy.from_dict(example_linear_threshold_stopping_policy.to_dict()).to_dict() ==
                example_linear_threshold_stopping_policy.to_dict())
        assert (LinearThresholdStoppingPolicy.from_dict(example_linear_threshold_stopping_policy.to_dict()) ==
                example_linear_threshold_stopping_policy)

    def test_linear_tabular_policy(self, example_linear_tabular_policy: LinearTabularPolicy) -> None:
        """
        Tests creation and dict conversion of the LinearTabularPolicy DAO

        :param example_linear_tabular_policy: an example LinearTabularPolicy
        :return: None
        """
        assert isinstance(example_linear_tabular_policy.to_dict(), dict)
        assert isinstance(LinearTabularPolicy.from_dict(example_linear_tabular_policy.to_dict()), LinearTabularPolicy)
        assert (LinearTabularPolicy.from_dict(example_linear_tabular_policy.to_dict()).to_dict() ==
                example_linear_tabular_policy.to_dict())
        assert LinearTabularPolicy.from_dict(example_linear_tabular_policy.to_dict()) == example_linear_tabular_policy

    def test_mixed_linear_tabular(self, example_mixed_linear_tabular: MixedLinearTabularPolicy) -> None:
        """
        Tests creation and dict conversion of the MixedLinearTabularPolicy DAO

        :param example_mixed_linear_tabular: an example MixedLinearTabularPolicy
        :return: None
        """
        assert isinstance(example_mixed_linear_tabular.to_dict(), dict)
        assert isinstance(MixedLinearTabularPolicy.from_dict(example_mixed_linear_tabular.to_dict()),
                          MixedLinearTabularPolicy)
        assert (MixedLinearTabularPolicy.from_dict(example_mixed_linear_tabular.to_dict()).to_dict() ==
                example_mixed_linear_tabular.to_dict())
        assert (MixedLinearTabularPolicy.from_dict(example_mixed_linear_tabular.to_dict()) ==
                example_mixed_linear_tabular)

    def test_mixed_multi_threshold_stopping_policy(
            self, example_mixed_multi_threshold_stopping_policy: MixedMultiThresholdStoppingPolicy) -> None:
        """
        Tests creation and dict conversion of the MixedMultiThresholdStoppingPolicy DAO

        :param example_mixed_multi_threshold_stopping_policy: an example MixedMultiThresholdStoppingPolicy
        :return: None
        """
        assert isinstance(example_mixed_multi_threshold_stopping_policy.to_dict(), dict)
        assert isinstance(MixedMultiThresholdStoppingPolicy.from_dict(
            example_mixed_multi_threshold_stopping_policy.to_dict()),
            MixedMultiThresholdStoppingPolicy)
        assert (MixedMultiThresholdStoppingPolicy.from_dict(example_mixed_multi_threshold_stopping_policy.to_dict()) ==
                example_mixed_multi_threshold_stopping_policy)

    def test_mixed_ppo_policy(self, example_mixed_ppo_policy: MixedPPOPolicy) -> None:
        """
        Tests creation and dict conversion of the MixedPPOPolicy DAO

        :param example_mixed_ppo_policy: an example MixedPPOPolicy
        :return: None
        """
        assert isinstance(example_mixed_ppo_policy.to_dict(), dict)
        assert isinstance(MixedPPOPolicy.from_dict(example_mixed_ppo_policy.to_dict()),
                          MixedPPOPolicy)
        assert MixedPPOPolicy.from_dict(example_mixed_ppo_policy.to_dict()).to_dict() == \
               example_mixed_ppo_policy.to_dict()
        assert MixedPPOPolicy.from_dict(example_mixed_ppo_policy.to_dict()) == example_mixed_ppo_policy

    def test_multi_threshold_stopping_policy(
            self, example_multi_threshold_stopping_policy: MultiThresholdStoppingPolicy) -> None:
        """
        Tests creation and dict conversion of the MultiThresholdStoppingPolicy DAO

        :param example_multi_threshold_stopping_policy: an example MultiThresholdStoppingPolicy
        :return: None
        """
        assert isinstance(example_multi_threshold_stopping_policy.to_dict(), dict)
        assert isinstance(MultiThresholdStoppingPolicy.from_dict(example_multi_threshold_stopping_policy.to_dict()),
                          MultiThresholdStoppingPolicy)
        assert (MultiThresholdStoppingPolicy.from_dict(example_multi_threshold_stopping_policy.to_dict()) ==
                example_multi_threshold_stopping_policy)

    def test_ppo_policy(self, example_ppo_policy: PPOPolicy) -> None:
        """
        Tests creation and dict conversion of the PPOPolicy DAO

        :param example_ppo_policy: an example PPOPolicy
        :return: None
        """
        assert isinstance(example_ppo_policy.to_dict(), dict)
        assert isinstance(PPOPolicy.from_dict(example_ppo_policy.to_dict()), PPOPolicy)
        assert PPOPolicy.from_dict(example_ppo_policy.to_dict()).to_dict() == example_ppo_policy.to_dict()
        assert PPOPolicy.from_dict(example_ppo_policy.to_dict()) == example_ppo_policy

    def test_random_policy(self, example_random_policy: RandomPolicy) -> None:
        """
        Tests creation and dict conversion of the RandomPolicy DAO

        :param example_random_policy: an example example_random_policy
        :return: None
        """
        assert isinstance(example_random_policy.to_dict(), dict)
        assert isinstance(RandomPolicy.from_dict(example_random_policy.to_dict()), RandomPolicy)
        assert RandomPolicy.from_dict(example_random_policy.to_dict()).to_dict() == example_random_policy.to_dict()
        assert RandomPolicy.from_dict(example_random_policy.to_dict()) == example_random_policy

    def test_vector_policy(self, example_vector_policy: VectorPolicy) -> None:
        """
        Tests creation and dict conversion of the VectorPolicy DAO

        :param example_vector_policy: an example VectorPolicy
        :return: None
        """
        assert isinstance(example_vector_policy.to_dict(), dict)
        assert isinstance(VectorPolicy.from_dict(example_vector_policy.to_dict()), VectorPolicy)
        assert VectorPolicy.from_dict(example_vector_policy.to_dict()).to_dict() == example_vector_policy.to_dict()
        assert VectorPolicy.from_dict(example_vector_policy.to_dict()) == example_vector_policy



import multiprocessing.pool


class NoDaemonProcess(multiprocessing.Process):
    """
    A process with the daemon property set to false
    """

    @property
    def daemon(self) -> bool:
        """
        The deamon property

        :return: True if the process is a Deamon, else false
        """
        return False

    @daemon.setter
    def daemon(self, value: bool) -> None:
        """
        The deamon setting

        :param value: the value to set the daemon property with
        :return: None
        """
        pass


class NoDaemonContext(type(multiprocessing.get_context())):  # type: ignore
    """
    Context of a non-daemon process
    """
    Process = NoDaemonProcess


class NestablePool(multiprocessing.pool.Pool):
    """
    Multiprocessing pool that allows nested multiprocessing
    """

    def __init__(self, *args, **kwargs) -> None:
        """
        Initializes the pool

        :param args: the pool arguments
        :param kwargs: the pool keyword arguments
        """
        kwargs['context'] = NoDaemonContext()
        super(NestablePool, self).__init__(*args, **kwargs)
